#!/usr/bin/env python

import logging
import os
import re
import sys
from pathlib import Path
from ruamel.yaml import YAML

from ansible.errors import AnsibleParserError
from ansible.parsing.dataloader import DataLoader
from ansible.parsing.mod_args import ModuleArgsParser
from ansible.parsing.yaml.objects import AnsibleMapping, AnsibleSequence

if os.environ.get("LSR_DEBUG") == "true":
    logging.getLogger().setLevel(logging.DEBUG)


ROLE_DIRS = [
    "defaults",
    "examples",
    "files",
    "handlers",
    "library",
    "meta",
    "module_utils",
    "tasks",
    "templates",
    "tests",
    "vars",
]

PLAY_KEYS = {
    "gather_facts",
    "handlers",
    "hosts",
    "import_playbook",
    "post_tasks",
    "pre_tasks",
    "roles",
    "tasks",
}

TASK_LIST_KWS = [
    "always",
    "block",
    "handlers",
    "post_tasks",
    "pre_tasks",
    "rescue",
    "tasks",
]


class LSRException(Exception):
    pass


def get_role_dir(role_path, dirpath):
    if role_path == dirpath:
        return None, None
    dir_pth = Path(dirpath)
    relpath = dir_pth.relative_to(role_path)
    base_dir = relpath.parts[0]
    if base_dir in ROLE_DIRS:
        return base_dir, relpath
    return None, None


def get_file_type(item):
    if isinstance(item, AnsibleMapping):
        if "galaxy_info" in item or "dependencies" in item:
            return "meta"
        return "vars"
    elif isinstance(item, AnsibleSequence):
        return "tasks"
    else:
        raise LSRException(f"Error: unknown type of file: {item}")


def get_item_type(item):
    if isinstance(item, AnsibleMapping):
        for key in PLAY_KEYS:
            if key in item:
                return "play"
        if "block" in item:
            return "block"
        return "task"
    else:
        raise LSRException(f"Error: unknown type of item: {item}")


class LSRFileTransformerBase(object):

    # we used to try to not deindent comment lines in the Ansible yaml,
    # but this changed the indentation when comments were used in
    # literal strings, which caused test failures - so for now, we
    # have to live with poorly indented Ansible comments . . .
    # INDENT_RE = re.compile(r'^  (?! *#)', flags=re.MULTILINE)
    INDENT_RE = re.compile(r"^  ", flags=re.MULTILINE)
    HEADER_RE = re.compile(r"^(---\n|.*\n---\n)", flags=re.DOTALL)
    FOOTER_RE = re.compile(r"\n([.][.][.]|[.][.][.]\n.*)$", flags=re.DOTALL)

    def __init__(self, filepath, rolename):
        self.filepath = filepath
        dl = DataLoader()
        self.ans_data = dl.load_from_file(filepath)
        if self.ans_data is None:
            raise LSRException(f"file is empty {filepath}")
        self.file_type = get_file_type(self.ans_data)
        self.rolename = rolename
        buf = open(filepath).read()
        self.ruamel_yaml = YAML(typ="rt")
        match = re.search(LSRFileTransformerBase.HEADER_RE, buf)
        if match:
            self.header = match.group(1)
        else:
            self.header = ""
        match = re.search(LSRFileTransformerBase.FOOTER_RE, buf)
        if match:
            self.footer = match.group(1) + "\n"
        else:
            self.footer = ""
        self.ruamel_yaml.default_flow_style = False
        self.ruamel_yaml.preserve_quotes = True
        self.ruamel_yaml.width = 1024
        self.ruamel_data = self.ruamel_yaml.load(buf)
        self.ruamel_yaml.indent(mapping=2, sequence=4, offset=2)
        self.outputfile = None
        self.outputstream = sys.stdout

    def run(self):
        if self.file_type == "vars":
            self.handle_vars(self.ans_data, self.ruamel_data)
        elif self.file_type == "meta":
            self.handle_meta(self.ans_data, self.ruamel_data)
        else:
            for a_item, ru_item in zip(self.ans_data, self.ruamel_data):
                self.handle_item(a_item, ru_item)

    def write(self):
        def xform(thing):
            logging.debug(f"xform thing {thing}")
            if self.file_type == "tasks":
                thing = re.sub(LSRFileTransformerBase.INDENT_RE, "", thing)
            thing = self.header + thing
            if not thing.endswith("\n"):
                thing = thing + "\n"
            thing = thing + self.footer
            return thing

        if self.outputfile:
            outstrm = open(self.outputfile, "w")
        else:
            outstrm = self.outputstream
        self.ruamel_yaml.dump(self.ruamel_data, outstrm, transform=xform)

    def task_cb(self, a_task, ru_task, module_name, module_args, delegate_to):
        """subclass will override"""
        pass

    def other_cb(self, a_item, ru_item):
        """subclass will override"""
        pass

    def vars_cb(self, a_item, ru_item):
        """subclass will override"""
        pass

    def meta_cb(self, a_item, ru_item):
        """subclass will override"""
        pass

    def handle_item(self, a_item, ru_item):
        """handle any type of item - call the appropriate handlers"""
        ans_type = get_item_type(a_item)
        self.handle_vars(a_item, ru_item)
        self.handle_other(a_item, ru_item)
        if ans_type == "task":
            self.handle_task(a_item, ru_item)
        self.handle_task_list(a_item, ru_item)

    def handle_other(self, a_item, ru_item):
        """handle properties of Ansible item other than vars and tasks"""
        self.other_cb(a_item, ru_item)

    def handle_vars(self, a_item, ru_item):
        """handle vars of Ansible item"""
        self.vars_cb(a_item, ru_item)

    def handle_meta(self, a_item, ru_item):
        """handle meta/main.yml file"""
        self.meta_cb(a_item, ru_item)

    def handle_task(self, a_task, ru_task):
        """handle a single task"""
        mod_arg_parser = ModuleArgsParser(a_task)
        try:
            action, args, delegate_to = mod_arg_parser.parse(
                skip_action_validation=True
            )
        except AnsibleParserError as e:
            raise LSRException(
                "Couldn't parse task at %s (%s)\n%s"
                % (a_task.ansible_pos, e.message, a_task)
            )
        self.task_cb(a_task, ru_task, action, args, delegate_to)

    def handle_task_list(self, a_item, ru_item):
        """item has one or more fields which hold a list of Task objects"""
        for kw in TASK_LIST_KWS:
            if kw in a_item:
                for a_task, ru_task in zip(a_item[kw], ru_item[kw]):
                    self.handle_item(a_task, ru_task)


def get_role_modules(role_path):
    """get the modules from the role
    returns a set() of module names"""
    role_modules = set()
    library_path = Path(os.path.join(role_path, "library"))
    if library_path.is_dir():
        for mod_file in library_path.iterdir():
            if mod_file.is_file() and mod_file.stem != "__init__":
                role_modules.add(mod_file.stem)
    return role_modules


class LSRTransformer(object):
    """Transform all of the .yml files in a role or role subdir"""

    def __init__(
        self,
        role_path,
        is_role_dir=True,
        role_name=None,
        file_xfrm_cls=LSRFileTransformerBase,
    ):
        """Create a role transformer.  The user can specify the specific class
        to use for transforming each file, and the extra arguments to pass to the
        constructor of that class
        is_role_dir - if True, role_path is the role directory (with all of the usual role subdirs)
                      if False, just operate on the .yml files found in role_path"""
        self.role_name = role_name
        self.role_path = role_path
        self.is_role_dir = is_role_dir
        self.file_xfrm_cls = file_xfrm_cls
        if self.is_role_dir and not self.role_name:
            self.role_name = os.path.basename(self.role_path)

    def run(self):
        for (dirpath, _, filenames) in os.walk(self.role_path):
            if self.is_role_dir:
                role_dir, _ = get_role_dir(self.role_path, dirpath)
                if not role_dir:
                    continue
            for filename in filenames:
                if not filename.endswith(".yml"):
                    continue
                filepath = os.path.join(dirpath, filename)
                logging.debug(f"filepath {filepath}")
                try:
                    lsrft = self.file_xfrm_cls(filepath, self.role_name)
                    lsrft.run()
                    lsrft.write()
                except LSRException as lsrex:
                    logging.debug(f"Could not transform {filepath}: {lsrex}")


if __name__ == "__main__":
    for role_path in sys.argv[1:]:
        lsrxfrm = LSRTransformer(role_path)
        lsrxfrm.run()



#!/usr/bin/python3

# Changes namespace, name, and version in Galaxy metadata.
# Useful for releasing to Automation Hub, where Collections live
# in namespaces separated from Ansible Galaxy.

import sys
from ruamel.yaml import YAML

filepath = "galaxy.yml"

buf = open(filepath).read()

yaml = YAML(typ="rt")
yaml.default_flow_style = False
yaml.preserve_quotes = True
yaml.indent(mapping=2, sequence=4, offset=2)

code = yaml.load(buf)

code["namespace"] = sys.argv[1]
code["name"] = sys.argv[2]
code["version"] = sys.argv[3]
code["description"] = sys.argv[4]
yaml.dump(code, sys.stdout)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright: (c) 2020, Red Hat, Inc.
# SPDX-License-Identifier: MIT

import argparse
import os
import shutil
import subprocess
import sys
import tempfile

import yaml

DEFAULT_GIT_SITE = "https://github.com"
DEFAULT_GIT_ORG = "linux-system-roles"


def build_collection(src_path, dest_path, galaxy, coll_rel, force):
    """
    Build a collection directory structure and the file.

    Convert the roles in src_path into a collection directory
    structure rooted at dest_path using the galaxy.yml metadata
    in galaxy and the collection_release metadata in coll_rel.
    The collection file for publishing will be in dest_path.
    """
    coll_dir = os.path.join(
        dest_path, "ansible_collections", galaxy["namespace"], galaxy["name"]
    )
    if os.path.isdir(coll_dir):
        if force:
            shutil.rmtree(coll_dir)
        else:
            raise Exception(
                "collection dest_path {} already exists - remove or use --force".format(
                    coll_dir
                )
            )
    os.makedirs(coll_dir, exist_ok=True)
    ignore_file_dir = os.path.join(coll_dir, "tests", "sanity")
    ignore_file = os.path.join(ignore_file_dir, "ignore-2.9.txt")
    collection_readme = os.path.join("lsr_role2collection", "collection_readme.md")
    collection_requirements = os.path.join(
        "lsr_role2collection", "collection_requirements.txt"
    )
    collection_requirements_dest = os.path.join(coll_dir, "requirements.txt")
    collection_bindep = os.path.join("lsr_role2collection", "collection_bindep.txt")
    collection_bindep_dest = os.path.join(coll_dir, "bindep.txt")
    ignore_file_src = os.path.join("lsr_role2collection", "extra-ignore-2.9.txt")
    ansible_lint = os.path.join("lsr_role2collection", ".ansible-lint")
    for rolename, roledata in coll_rel.items():
        if rolename.startswith("_"):
            continue
        roledir = os.path.join(src_path, rolename)
        if os.path.isdir(roledir):
            subprocess.check_call(["git", "fetch"], cwd=roledir)
        else:
            git_url = "{}/{}/{}".format(
                DEFAULT_GIT_SITE,
                roledata.get("org", DEFAULT_GIT_ORG),
                roledata.get("repo", rolename),
            )
            subprocess.check_call(
                [
                    "git",
                    "-c",
                    "advice.detachedHead=false",
                    "clone",
                    "-q",
                    git_url,
                    roledir,
                ]
            )
        subprocess.check_call(
            ["git", "-c", "advice.detachedHead=false", "checkout", roledata["ref"]],
            cwd=roledir,
        )
        subrole_prefix = f"private_{rolename}_subrole_"
        subprocess.check_call(
            [
                "python",
                "lsr_role2collection.py",
                "--src-owner",
                DEFAULT_GIT_ORG,
                "--role",
                rolename,
                "--src-path",
                src_path,
                "--dest-path",
                dest_path,
                "--namespace",
                galaxy["namespace"],
                "--collection",
                galaxy["name"],
                "--subrole-prefix",
                subrole_prefix,
                "--readme",
                collection_readme,
            ]
        )
        role_ignore_file = os.path.join(roledir, ".sanity-ansible-ignore-2.9.txt")
        if os.path.isfile(role_ignore_file):
            if not os.path.isdir(ignore_file_dir):
                os.mkdir(ignore_file_dir)
            with open(ignore_file, "a") as ign_fd:
                with open(role_ignore_file, "r") as role_ign_fd:
                    ign_fd.write(role_ign_fd.read())
    shutil.copy(galaxy["_filename"], coll_dir)
    shutil.copy(coll_rel["_filename"], coll_dir)
    if os.path.exists(collection_requirements):
        shutil.copy(collection_requirements, collection_requirements_dest)
    if os.path.exists(collection_bindep):
        shutil.copy(collection_bindep, collection_bindep_dest)
    # removing dot files/dirs
    subprocess.check_call(["bash", "-c", f"rm -r {coll_dir}/.[A-Za-z]*"])
    # copy required dot files like .ansible-lint here
    shutil.copy(ansible_lint, coll_dir)
    with open(ignore_file, "a") as ign_fd:
        with open(ignore_file_src, "r") as role_ign_fd:
            ign_fd.write(role_ign_fd.read())

    build_args = ["ansible-galaxy", "collection", "build", "-v"]
    if force:
        build_args.append("-f")
    build_args.append(coll_dir)
    subprocess.check_call(build_args, cwd=dest_path)


def check_collection(dest_path, galaxy):
    coll_file = (
        f"{dest_path}/{galaxy['namespace']}-{galaxy['name']}-{galaxy['version']}.tar.gz"
    )
    gi_config = "lsr_role2collection/galaxy-importer.cfg"
    subprocess.check_call(
        ["python", "-m", "galaxy_importer.main", coll_file],
        cwd=dest_path,
        env={"GALAXY_IMPORTER_CONFIG": gi_config},
    )


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--galaxy-yml",
        type=open,
        default=os.environ.get("GALAXY_YML", "galaxy.yml"),
        help="Path/filename for galaxy.yml",
    )
    parser.add_argument(
        "--collection-release-yml",
        type=open,
        default=os.environ.get("COLLECTION_RELEASE_YML", "collection_release.yml"),
        help="Path/filename for collection_release.yml",
    )
    parser.add_argument(
        "--src-path",
        type=str,
        default=os.environ.get("COLLECTION_SRC_PATH"),
        help="Path to the directory containing the local clone of the role repos",
    )
    parser.add_argument(
        "--dest-path",
        type=str,
        default=os.environ.get(
            "COLLECTION_DEST_PATH", os.environ.get("HOME") + "/.ansible/collections"
        ),
        help=(
            "Path to collection base directory; collection directory structure "
            "will be created in DIR/ansible_collection/NAMESPACE/COLLECTION_NAME; "
            "collection package file will be created in DIR; "
            "default to ${HOME}/.ansible/collections"
        ),
    )
    parser.add_argument(
        "--force",
        default=False,
        action="store_true",
        help="Remove collection destination dir and file before creating",
    )
    args = parser.parse_args()

    galaxy = yaml.safe_load(args.galaxy_yml)
    galaxy["_filename"] = args.galaxy_yml.name
    coll_rel = yaml.safe_load(args.collection_release_yml)
    coll_rel["_filename"] = args.collection_release_yml.name
    workdir = None
    if not args.src_path:
        workdir = tempfile.mkdtemp(suffix=".lsr", prefix="collection")
        args.src_path = workdir
    try:
        build_collection(args.src_path, args.dest_path, galaxy, coll_rel, args.force)
        check_collection(args.dest_path, galaxy)
    finally:
        if workdir:
            shutil.rmtree(workdir)


if __name__ == "__main__":
    sys.exit(main())



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright: (c) 2021, Red Hat, Inc.
# SPDX-License-Identifier: MIT

import argparse
import os
import re
import shutil
import subprocess
import sys
import tempfile

import yaml

DEFAULT_GIT_SITE = "https://github.com"
DEFAULT_GIT_ORG = "linux-system-roles"


def update_collection(src_path, coll_rel, use_commit_hash):
    """
    Update refs in collection_release.yml.

    Use the latest tag for the ref.  If use_commit_hash is True
    and the latest commit is not tagged, use the commit hash of
    the latest commit for the ref.
    """
    for rolename in coll_rel:
        if rolename.startswith("_"):
            continue
        roledir = os.path.join(src_path, rolename)
        if os.path.isdir(roledir):
            subprocess.check_call(["git", "fetch"], cwd=roledir)
        else:
            git_url = "{}/{}/{}".format(
                DEFAULT_GIT_SITE,
                coll_rel[rolename].get("org", DEFAULT_GIT_ORG),
                coll_rel[rolename].get("repo", rolename),
            )
            subprocess.check_call(
                [
                    "git",
                    "-c",
                    "advice.detachedHead=false",
                    "clone",
                    "-q",
                    git_url,
                    roledir,
                ]
            )
        branch_output = subprocess.run(
            ["git", "branch", "-r"],
            cwd=roledir,
            encoding="utf-8",
            stdout=subprocess.PIPE,
        )
        if branch_output.returncode != 0:
            print(
                f"ERROR: git branch -r failed - {branch_output.stdout} {branch_output.stderr}"
            )
        mmatch = re.search(r"origin/HEAD -> origin/(\w+)", branch_output.stdout)
        main_branch = mmatch.group(1)
        subprocess.check_call(
            ["bash", "-c", f"git checkout {main_branch}; git pull"],
            cwd=roledir,
        )
        describe_cmd = ["git", "describe", "--tags"]
        describe_output = subprocess.run(
            describe_cmd, cwd=roledir, encoding="utf-8", stdout=subprocess.PIPE
        )
        if describe_output.returncode != 0:
            print(
                f"ERROR: {describe_cmd} failed - {describe_output.stdout} {describe_output.stderr}"
            )
        describe_ary = describe_output.stdout.strip().split("-")
        if len(describe_ary) == 1 or not use_commit_hash:
            gitref = describe_ary[0]  # tag
        else:
            gitref = describe_ary[2][1:]  # commit hash - skip leading "g"
        coll_rel[rolename]["ref"] = gitref


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--collection-release-yml",
        type=open,
        default=os.environ.get("COLLECTION_RELEASE_YML", "collection_release.yml"),
        help="Path/filename for collection_release.yml",
    )
    parser.add_argument(
        "--src-path",
        type=str,
        default=os.environ.get("COLLECTION_SRC_PATH"),
        help="Path to the directory containing the local clone of the role repos",
    )
    parser.add_argument(
        "--use-commit-hash",
        default=False,
        action="store_true",
        help="Only use version tags for refs - if True, use latest HEAD commit hash if not tagged",
    )
    args = parser.parse_args()

    coll_rel = yaml.safe_load(args.collection_release_yml)
    coll_rel["_filename"] = args.collection_release_yml.name
    workdir = None
    if not args.src_path:
        workdir = tempfile.mkdtemp(suffix=".lsr", prefix="collection")
        args.src_path = workdir
    try:
        update_collection(args.src_path, coll_rel, args.use_commit_hash)
        with open(coll_rel["_filename"], "w") as crf:
            del coll_rel["_filename"]
            yaml.dump(coll_rel, crf)
    finally:
        if workdir:
            shutil.rmtree(workdir)


if __name__ == "__main__":
    sys.exit(main())



#!/usr/bin/python3 -u
"""Check PR statuses for linux-system-roles github"""

import argparse
import datetime
import itertools
import os
import re
import sys
import logging
import yaml
from collections import OrderedDict
from ghapi.all import GhApi
from ghapi.page import paged as gh_paged
from ghapi.core import gh2date
from operator import attrgetter

DEFAULT_EXCLUDES = [
    ".github",
    "auto-maintenance",
    "ci-testing",
    "experimental-azure-firstboot",
    "image_builder",
    "linux-system-roles.github.io",
    "lsr-gh-action-py26",
    "meta_test",
    "sap-base-settings",
    "sap-hana-preconfigure",
    "sap-netweaver-preconfigure",
    "sap-preconfigure",
    "test-harness",
    "tox-lsr",
    "tuned",
]

DEFAULT_ANSIBLES = ["ansible-2.8", "ansible-2.9"]

DEFAULT_PLATFORMS = [
    "centos-6",
    "centos-7",
    "centos-8",
    "fedora-32",
    "fedora-33",
    "fedora-34",
    "rhel-6",
    "rhel-7",
    "rhel-8-y",
    "rhel-8",
    "rhel-x",
]

DEFAULT_STATES = ["pending", "failure", "success", "abandoned", "error"]

DEFAULT_ENVIRONMENTS = ["staging", "production"]


def match_environment(env, status, args):
    if status.context.startswith("linux-system-roles-test/"):
        env = "production"
    elif status.context.startswith("linux-system-roles-test-staging/"):
        env = "staging"
    elif env.endswith(" (staging)"):
        env = "staging"
    else:
        env = "production"
    return env in args.env


def match_ansible(ansible, args):
    return ansible in args.ansible


def match_platform(platform, args):
    return platform in args.platform


def match_user(user, args):
    if not args.user:
        return True
    return user in args.user


def conv_to_aware(dt, tz="+00:00"):
    """Convert a naive dt object to an aware dt object using the given string offset timezone"""
    if isinstance(dt, datetime.datetime):
        if dt.microsecond:
            return datetime.datetime.strptime(str(dt) + tz, "%Y-%m-%d %H:%M:%S.%f%z")
        else:
            return datetime.datetime.strptime(str(dt) + tz, "%Y-%m-%d %H:%M:%S%z")
    else:
        return gh2date(dt.replace("Z", "+00:00"))


datetime_min = conv_to_aware(datetime.datetime.min)
datetime_max = conv_to_aware(datetime.datetime.max)


def gh_iter(op, subfield, *args, **kwargs):
    """Hide paging iterator details from callers."""
    for attr in ["page", "per_page"]:
        if attr not in op.params:
            op.params.append(attr)
    pages = gh_paged(op, *args, **kwargs)
    for page in pages:
        if subfield:
            items = getattr(page, subfield)
            if not items:
                pages.close()
            for subitem in items:
                yield subitem
        else:
            for item in page:
                yield item


class Summary(OrderedDict):
    def __init__(self, **kwargs):
        for k, v in kwargs.items():
            setattr(self, k, v)

    def __getattr__(self, key):
        self.__setattr__(key, 0)
        return 0


def get_statuses(gh, args, repo, pr, updated_since):
    excluded_statuses = set()
    statuses = Summary(oldest=datetime_max, newest=datetime_min)
    for status in gh_iter(
        gh.repos.get_combined_status_for_ref,
        "statuses",
        owner=args.org,
        repo=repo.name,
        ref=pr.head.sha,
        per_page=99,
    ):
        logging.debug(
            f"\t\t\t{status.state} {status.context} {status.updated_at} {status.description}"
        )
        updated_at_conv = conv_to_aware(status.updated_at)
        seen_status = statuses.get(status.context)
        if seen_status is None or seen_status.updated_at < updated_at_conv:
            statuses.count += 1
            # e.g. fedora-33/ansible-2.9 OR fedora-33/ansible-2.9 (staging)
            match = re.match(r"([^/]+)/(\S+)(.*)", status.context)
            if not match or not len(match.groups()) == 3:
                logging.debug(f"unknown context {status.context}")
                continue
            platform, ansible, environment = match.groups()
            if not ansible.startswith("ansible-"):
                logging.debug(f"unknown context {status.context}")
                continue
            if (
                status.state in args.state
                and match_environment(environment, status, args)
                and match_ansible(ansible, args)
                and match_platform(platform, args)
                and (not args.updated_since or updated_at_conv > updated_since)
            ):
                setattr(status, "updated_at_conv", updated_at_conv)
                setattr(status, "ansible", ansible)
                setattr(status, "platform", platform)
                setattr(status, "env", environment)
                statuses[status.context] = status
                if status.state == "pending":
                    statuses.pending += 1
                elif status.state != "success":
                    statuses.failed += 1
                if statuses.oldest > updated_at_conv:
                    statuses.oldest = updated_at_conv
                if statuses.newest < updated_at_conv:
                    statuses.newest = updated_at_conv
            else:
                excluded_statuses.add(status.context)
    return statuses, excluded_statuses


def get_checks(gh, args, repo, pr):
    checks = Summary(
        oldest_start=datetime_max,
        newest_start=datetime_min,
        oldest_finish=datetime_max,
        newest_finish=datetime_min,
    )
    for check_run in gh_iter(
        gh.checks.list_for_ref,
        "check_runs",
        owner=args.org,
        repo=repo.name,
        ref=pr.head.sha,
        per_page=99,
    ):
        checks.count += 1
        setattr(check_run, "started_at_conv", conv_to_aware(check_run.started_at))
        if checks.oldest_start > check_run.started_at_conv:
            checks.oldest_start = check_run.started_at_conv
        if checks.newest_start < check_run.started_at_conv:
            checks.newest_start = check_run.started_at_conv
        if check_run.status != "completed":
            checks.pending += 1
        else:
            setattr(
                check_run, "completed_at_conv", conv_to_aware(check_run.completed_at)
            )
            if checks.oldest_finish > check_run.completed_at_conv:
                checks.oldest_finish = check_run.completed_at_conv
            if checks.newest_finish < check_run.completed_at_conv:
                checks.newest_finish = check_run.completed_at_conv
            if check_run.conclusion not in ["success", "neutral"]:
                checks.failed += 1
        checks[check_run.name] = check_run
    return checks


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--token",
        help="github oauth token",
    )
    parser.add_argument(
        "--exclude",
        default=DEFAULT_EXCLUDES,
        action="append",
        help="repos to exclude",
    )
    parser.add_argument(
        "--include",
        default=[],
        action="append",
        help="repos to include",
    )
    parser.add_argument(
        "--org",
        default="linux-system-roles",
        help="github organization",
    )
    parser.add_argument(
        "--state",
        action="append",
        help="only show statuses that have these states",
    )
    parser.add_argument(
        "--env",
        action="append",
        help="only show statuses from these environments",
    )
    parser.add_argument(
        "--ansible",
        action="append",
        help="only show statuses from tests with these versions of ansible",
    )
    parser.add_argument(
        "--platform",
        action="append",
        help="only show statuses from tests with these platforms",
    )
    parser.add_argument(
        "-v",
        "--verbose",
        action="count",
        default=0,
        help="-v -v to increase verbosity",
    )
    parser.add_argument(
        "--user",
        action="append",
        help="only show statuses from PRs submitted by these users",
    )
    parser.add_argument(
        "--updated-since",
        action="store",
        default=0,
        type=int,
        help="only show statuses updated since this many hours ago",
    )
    parser.add_argument(
        "--sort-spec",
        action="append",
        default=[],
        help="sort PRs by these criteria",
    )
    parser.add_argument(
        "--stat-type",
        choices=["status", "check", "both"],
        default="both",
        help="Return only statuses or checks",
    )

    args = parser.parse_args()
    if not args.state:
        args.state = DEFAULT_STATES
    if not args.env:
        args.env = DEFAULT_ENVIRONMENTS
    if not args.ansible:
        args.ansible = DEFAULT_ANSIBLES
    if not args.platform:
        args.platform = DEFAULT_PLATFORMS
    if not args.token:
        with open(f"{os.environ['HOME']}/.config/hub") as hub:
            hsh = yaml.safe_load(hub)
            args.token = hsh["github.com"][0]["oauth_token"]
    no_status_error_time = datetime.timedelta(hours=8)
    pending_error_time = datetime.timedelta(hours=8)
    status_env = ["" if xx == "production" else " (staging)" for xx in args.env]
    required_statuses = set(
        [
            f"{platform}/{ansible}{env}"
            for platform, ansible, env in itertools.product(
                args.platform, args.ansible, status_env
            )
            if not (platform.startswith("fedora-") and ansible.endswith("-2.8"))
        ]
    )

    prs = []
    gh = GhApi(token=args.token)
    rate_limit = gh.rate_limit.get()
    print(f"github limit remaining is {rate_limit.rate.remaining}")
    now = datetime.datetime.now(tz=datetime.timezone.utc)
    logging.debug(f"time now is {now}")

    # gather the data
    for repo in gh_iter(gh.repos.list_for_org, "", args.org, per_page=100):
        if repo.name in args.exclude:
            logging.debug(f"Excluding repo {repo.name}")
            continue
        if args.include and repo.name not in args.include:
            logging.debug(f"repo not included {repo.name}")
            continue
        for pr in gh_iter(
            gh.pulls.list,
            "",
            owner=args.org,
            repo=repo.name,
            state="open",
            sort="created",
            per_page=100,
        ):
            if not match_user(pr.user.login, args):
                continue
            pr_updated_at = conv_to_aware(pr.updated_at)
            if args.updated_since:
                updated_since = now - datetime.timedelta(hours=args.updated_since)
            else:
                updated_since = None
            if args.stat_type in ["both", "status"]:
                statuses, excluded_statuses = get_statuses(
                    gh, args, repo, pr, updated_since
                )
                setattr(pr, "statuses", statuses)
                setattr(pr, "excluded_statuses", excluded_statuses)
                if args.stat_type == "status":
                    setattr(pr, "checks", [])
            if args.stat_type in ["both", "check"]:
                checks = get_checks(gh, args, repo, pr)
                setattr(pr, "checks", checks)
                if args.stat_type == "check":
                    setattr(pr, "statuses", [])
                    setattr(pr, "excluded_statuses", [])
            setattr(pr, "updated_at_conv", pr_updated_at)
            prs.append(pr)

    # sort the list by the field of pr
    # updated_at_conv
    # statuses.oldest
    # statuses.newest
    # checks.oldest_start
    # checks.newest_start
    # checks.oldest_finish
    # checks.newest_finish
    # statuses.failed
    # checks.failed
    # statuses.pending
    # checks.pending
    for raw_key in args.sort_spec:
        match = re.match(r"^(.+?)([-+])?$", raw_key)
        key = match.group(1)
        rev = match.group(2) == "-"
        prs.sort(key=attrgetter(key), reverse=rev)

    # report the data
    for pr in prs:
        indent = ""
        print(f"{indent}PR {pr.base.repo.name}#{pr.number} {pr.title}")
        indent += "  "
        print(f"{indent}{pr.user.login} Updated:{pr.updated_at} {pr.html_url}")
        checks = pr.checks
        statuses = pr.statuses
        excluded_statuses = pr.excluded_statuses
        if not checks:
            print(f"{indent}no checks for PR")
        elif checks.pending or checks.failed:
            print(
                f"{indent}checks: {checks.count} total - {checks.pending} pending - {checks.failed} failed"
            )
            if args.verbose:
                for name, check_run in checks.items():
                    print(
                        f"{indent}  {name} status {check_run.status} conclusion {check_run.conclusion} "
                        f"started_at {check_run.started_at} completed_at {check_run.completed_at}"
                    )
        else:
            print(f"{indent}checks: {checks.count} total all passed")

        if not statuses and now - pr_updated_at > no_status_error_time:
            print(f"{indent}problem - PR has no status after {no_status_error_time}")
        elif not statuses:
            print(f"{indent}no status for PR")
        else:
            for context in required_statuses:
                if context not in statuses and context not in excluded_statuses:
                    statuses.missing += 1
                    status = argparse.Namespace()
                    setattr(status, "state", "missing")
                    statuses[context] = status
            if statuses.pending or statuses.failed or statuses.missing:
                print(
                    f"{indent}statuses: matched {len(statuses)} out of {statuses.count} total "
                    f"- {statuses.pending} pending "
                    f"- {statuses.failed} failed - {statuses.missing} missing"
                )
                oldest_pending_status = None
                oldest_pending_status_dt = None
                for context, status in statuses.items():
                    if args.verbose:
                        if status.state == "missing":
                            print(f"{indent}  {context} is missing")
                            continue
                        print(
                            f"{indent}  {context} status {status.state} updated {status.updated_at}"
                        )
                    if status.state == "pending" and (
                        not oldest_pending_status_dt
                        or status.updated_at_conv < oldest_pending_status_dt
                    ):
                        oldest_pending_status_dt = status.updated_at_conv
                        oldest_pending_status = status
                if oldest_pending_status:
                    status_age = now - oldest_pending_status_dt
                    print(
                        f"{indent}oldest pending status is {oldest_pending_status_dt} for "
                        f"status {oldest_pending_status.context} diff {status_age}"
                    )
                    if status_age > pending_error_time:
                        print(
                            f"{indent}possible hang - age of {oldest_pending_status_dt} for "
                            f"status {oldest_pending_status.context}"
                        )
            else:
                print(
                    f"{indent}statuses: matched {len(statuses)} out of {statuses.count} total - all passed"
                )
    rate_limit = gh.rate_limit.get()
    print(f"github limit remaining is {rate_limit.rate.remaining}")


if __name__ == "__main__":
    sys.exit(main())



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# (c) 2020 Matt Martz <matt@sivel.net>
# GNU General Public License v3.0+
#     (see https://www.gnu.org/licenses/gpl-3.0.txt)

# Usage:
# lsr-role2collection.py [--namespace COLLECTION_NAMESPACE]
#                        [--collection COLLECTION_NAME]
#                        --src-path COLLECTION_SRC_PATH
#                        --dest-path COLLECTION_DEST_PATH
#                        --role ROLE_NAME
#                        [--subrole-prefix STR]
#                        [--replace-dot STR]
#                        [-h]
# Or
#
# COLLECTION_SRC_PATH=/path/to/{src_owner} \
# COLLECTION_DEST_PATH=/path/to/collections \
# COLLECTION_NAMESPACE=mynamespace \
# COLLECTION_NAME=myname \
# lsr-role2collection.py --role ROLE_NAME
#   ROLE_NAME role must exist in COLLECTION_SRC_PATH
#   Converted collections are placed in COLLECTION_DEST_PATH/ansible_collections/COLLECTION_NAMESPACE/COLLECTION_NAME

import argparse
import errno
import fnmatch
import logging
import os
import re
import sys
import textwrap

from pathlib import Path
from ruamel.yaml import YAML
from shutil import copytree, copy2, copyfile, ignore_patterns, rmtree
from operator import itemgetter

ALL_ROLE_DIRS = [
    "defaults",
    "examples",
    "files",
    "handlers",
    "library",
    "meta",
    "module_utils",
    "tasks",
    "templates",
    "tests",
    "vars",
]

PLAY_KEYS = {
    "gather_facts",
    "handlers",
    "hosts",
    "import_playbook",
    "post_tasks",
    "pre_tasks",
    "roles",
    "tasks",
}

TASK_LIST_KWS = [
    "always",
    "block",
    "handlers",
    "post_tasks",
    "pre_tasks",
    "rescue",
    "tasks",
]


class LSRException(Exception):
    pass


def get_role_dir(role_path, dirpath):
    dir_pth = Path(dirpath)
    if role_path == dir_pth:
        return None, None
    relpath = dir_pth.relative_to(role_path)
    base_dir = relpath.parts[0]
    if base_dir in ALL_ROLE_DIRS:
        return base_dir, relpath
    return None, None


def get_file_type(item):
    if isinstance(item, dict):
        if "galaxy_info" in item or "dependencies" in item:
            return "meta"
        return "vars"
    elif isinstance(item, list):
        return "tasks"
    else:
        raise LSRException(f"Error: unknown type of file: {item}")


def get_item_type(item):
    if isinstance(item, dict):
        for key in PLAY_KEYS:
            if key in item:
                return "play"
        if "block" in item:
            return "block"
        return "task"
    else:
        raise LSRException(f"Error: unknown type of item: {item}")


class LSRFileTransformerBase(object):

    # we used to try to not deindent comment lines in the Ansible yaml,
    # but this changed the indentation when comments were used in
    # literal strings, which caused test failures - so for now, we
    # have to live with poorly indented Ansible comments . . .
    # INDENT_RE = re.compile(r'^  (?! *#)', flags=re.MULTILINE)
    INDENT_RE = re.compile(r"^  ", flags=re.MULTILINE)
    HEADER_RE = re.compile(r"^(---\n|.*\n---\n)", flags=re.DOTALL)
    FOOTER_RE = re.compile(r"\n([.][.][.]|[.][.][.]\n.*)$", flags=re.DOTALL)

    def __init__(self, filepath, rolename, newrolename, args):
        self.filepath = filepath
        self.namespace = args["namespace"]
        self.collection = args["collection"]
        self.prefix = args["prefix"]
        self.subrole_prefix = args["subrole_prefix"]
        self.replace_dot = args["replace_dot"]
        self.rolename_regex = "[{0}.]".format(self.replace_dot)
        self.role_modules = args["role_modules"]
        self.src_owner = args["src_owner"]
        self.top_dir = args["top_dir"]
        self.rolename = rolename
        self.newrolename = newrolename
        self.extra_mapping_src_owner = args["extra_mapping_src_owner"]
        self.extra_mapping_src_role = args["extra_mapping_src_role"]
        self.extra_mapping_dest_prefix = args["extra_mapping_dest_prefix"]
        self.extra_mapping_dest_role = args["extra_mapping_dest_role"]
        buf = open(filepath, encoding="utf-8").read()
        self.ruamel_yaml = YAML(typ="rt")
        match = re.search(LSRFileTransformerBase.HEADER_RE, buf)
        if match:
            self.header = match.group(1)
        else:
            self.header = ""
        match = re.search(LSRFileTransformerBase.FOOTER_RE, buf)
        if match:
            self.footer = match.group(1) + "\n"
        else:
            self.footer = ""
        self.ruamel_yaml.default_flow_style = False
        self.ruamel_yaml.preserve_quotes = True
        self.ruamel_yaml.width = 1024
        self.ruamel_data = self.ruamel_yaml.load(buf)
        self.ruamel_yaml.indent(mapping=2, sequence=4, offset=2)
        self.file_type = get_file_type(self.ruamel_data)
        self.outputfile = None
        self.outputstream = sys.stdout

    def run(self):
        if self.file_type == "vars":
            self.handle_vars(self.ruamel_data)
        elif self.file_type == "meta":
            self.handle_meta(self.ruamel_data)
        else:
            for item in self.ruamel_data:
                self.handle_item(item)

    def write(self):
        def xform(thing):
            logging.debug(f"xform thing {thing}")
            if self.file_type == "tasks":
                thing = re.sub(LSRFileTransformerBase.INDENT_RE, "", thing)
            thing = self.header + thing
            if not thing.endswith("\n"):
                thing = thing + "\n"
            thing = thing + self.footer
            return thing

        if self.outputfile:
            outstrm = open(self.outputfile, "w", encoding="utf-8")
        else:
            outstrm = self.outputstream
        self.ruamel_yaml.dump(self.ruamel_data, outstrm, transform=xform)

    def task_cb(self, task):
        """subclass will override"""
        pass

    def other_cb(self, item):
        """subclass will override"""
        pass

    def vars_cb(self, item):
        """subclass will override"""
        pass

    def meta_cb(self, item):
        """subclass will override"""
        pass

    def handle_item(self, item):
        """handle any type of item - call the appropriate handlers"""
        ans_type = get_item_type(item)
        self.handle_vars(item)
        self.handle_other(item)
        if ans_type == "task":
            self.handle_task(item)
        self.handle_task_list(item)

    def handle_other(self, item):
        """handle properties of Ansible item other than vars and tasks"""
        self.other_cb(item)

    def handle_vars(self, item):
        """handle vars of Ansible item"""
        self.vars_cb(item)

    def handle_meta(self, item):
        """handle meta/main.yml file"""
        self.meta_cb(item)

    def handle_task(self, task):
        """handle a single task"""
        self.task_cb(task)

    def handle_task_list(self, item):
        """item has one or more fields which hold a list of Task objects"""
        for kw in TASK_LIST_KWS:
            if kw in item:
                for task in item[kw]:
                    self.handle_item(task)


def get_role_modules(role_path):
    """get the modules from the role
    returns a set() of module names"""
    role_modules = set()
    library_path = Path(os.path.join(role_path, "library"))
    if library_path.is_dir():
        for mod_file in library_path.iterdir():
            if mod_file.is_file() and mod_file.stem != "__init__":
                role_modules.add(mod_file.stem)
    return role_modules


class LSRTransformer(object):
    """Transform all of the .yml files in a role or role subdir"""

    def __init__(
        self,
        role_path,
        transformer_args,
        is_role_dir=True,
        role_name=None,
        new_role_name=None,
        file_xfrm_cls=LSRFileTransformerBase,
    ):
        """Create a role transformer.  The user can specify the specific class
        to use for transforming each file, and the extra arguments to pass to the
        constructor of that class
        is_role_dir - if True, role_path is the role directory (with all of the usual role subdirs)
                      if False, just operate on the .yml files found in role_path"""
        self.role_name = role_name
        self.new_role_name = new_role_name
        self.role_path = role_path
        self.is_role_dir = is_role_dir
        self.transformer_args = transformer_args
        self.file_xfrm_cls = file_xfrm_cls
        if self.is_role_dir and not self.role_name:
            self.role_name = os.path.basename(self.role_path)

    def run(self):
        for (dirpath, _, filenames) in os.walk(self.role_path):
            if self.is_role_dir:
                role_dir, _ = get_role_dir(self.role_path, dirpath)
                if not role_dir:
                    continue
            for filename in filenames:
                if not filename.endswith(".yml"):
                    continue
                filepath = os.path.join(dirpath, filename)
                logging.debug(f"filepath {filepath}")
                try:
                    lsrft = self.file_xfrm_cls(
                        filepath,
                        self.role_name,
                        self.new_role_name,
                        self.transformer_args,
                    )
                    lsrft.run()
                    lsrft.write()
                except LSRException as lsrex:
                    logging.debug(f"Could not transform {filepath}: {lsrex}")


# ==============================================================================

ROLE_DIRS = (
    "defaults",
    "files",
    "handlers",
    "meta",
    "tasks",
    "templates",
    "vars",
)

PLUGINS = (
    "action_plugins",
    "become_plugins",
    "cache_plugins",
    "callback_plugins",
    "cliconf_plugins",
    "connection_plugins",
    "doc_fragments",
    "filter_plugins",
    "httpapi_plugins",
    "inventory_plugins",
    "library",
    "lookup_plugins",
    "module_utils",
    "netconf_plugins",
    "shell_plugins",
    "strategy_plugins",
    "terminal_plugins",
    "test_plugins",
    "vars_plugins",
)

TESTS = ("tests",)

DOCS = (
    "docs",
    "design_docs",
    "examples",
    "README.md",
    "DCO",
)

TOX = (
    ".ansible-lint",
    ".flake8",
    ".pre-commit-config.yaml",
    ".pydocstyle",
    ".travis",
    ".travis.yml",
    ".yamllint_defaults.yml",
    ".yamllint.yml",
    ".yamllint.yaml",
    "ansible_pytest_extra_requirements.txt",
    "custom_requirements.txt",
    "molecule",
    "molecule_extra_requirements.txt",
    "pylintrc",
    "pylint_extra_requirements.txt",
    "pytest_extra_requirements.txt",
    "tox.ini",
    "tuned_requirements.txt",
)

DO_NOT_COPY = (
    ".github",
    ".gitignore",
    ".lgtm.yml",
    ".tox",
    ".venv",
    "artifacts",
    "run_pylint.py",
    "scripts",
    "semaphore",
    "standard-inventory-qcow2",
    "Vagrantfile",
    "README.html",
    "CHANGELOG",
)

ALL_DIRS = ROLE_DIRS + PLUGINS + TESTS + DOCS + DO_NOT_COPY

IMPORT_RE = re.compile(
    br"(\bimport) (ansible\.module_utils\.)(\S+)(.*)(\s+#.+|.*)$", flags=re.M
)
FROM_RE = re.compile(
    br"(\bfrom) (ansible\.module_utils\.?)(\S+)? import (\(*(?:\n|\r\n)?)(\S+)(\s+#.+|.*)$",
    flags=re.M,
)

if os.environ.get("LSR_DEBUG") == "true":
    logging.getLogger().setLevel(logging.DEBUG)
elif os.environ.get("LSR_INFO") == "true":
    logging.getLogger().setLevel(logging.INFO)
else:
    logging.getLogger().setLevel(logging.ERROR)


class LSRFileTransformer(LSRFileTransformerBase):
    """Do the role file transforms - fix role names, add FQCN
    to module names, etc."""

    def convert_rolename(self, rolename, lsr_rolename=None):
        """convert the given rolename to the new name"""
        if rolename.count(".") == 1:
            _src_owner, _rolename_base = rolename.split(".")
        else:
            _src_owner = None
            _rolename_base = rolename
        if not lsr_rolename:
            lsr_rolename = self.src_owner + "." + self.rolename
        logging.debug(f"\ttask role {rolename}")
        new_name = None
        if rolename == lsr_rolename or self.comp_rolenames(rolename, self.rolename):
            new_name = self.prefix + self.newrolename
        elif _rolename_base and _rolename_base in self.extra_mapping_src_role:
            _src_role_index = self.extra_mapping_src_role.index(_rolename_base)
            # --extra-mapping "SRC_OWNER0.SRC_ROLE0:DEST_PREFIX[.]DEST_ROLE1,
            #                  SRC_ROLE1:DEST_PREFIX[.]DEST_ROLE1"
            # current _rolename_base is SRC_ROLE0 and
            #   _src_owner is None or _src_owner is SRC_OWNER0
            # or current _rolename_base is SRC_ROLE1
            if (
                not _src_owner
                or (_src_owner == self.extra_mapping_src_owner[_src_role_index])
                or (
                    not self.extra_mapping_src_owner[_src_role_index]
                    and _src_owner == self.src_owner
                )
            ):
                new_name = "{0}{1}".format(
                    self.extra_mapping_dest_prefix[_src_role_index]
                    if self.extra_mapping_dest_prefix[_src_role_index]
                    else self.prefix,
                    self.extra_mapping_dest_role[_src_role_index],
                )
        elif rolename.startswith("{{ role_path }}"):
            match = re.match(r"{{ role_path }}/roles/([\w\d.]+)", rolename)
            if match.group(1).startswith(self.subrole_prefix):
                new_name = self.prefix + match.group(1).replace(".", self.replace_dot)
            else:
                new_name = (
                    self.prefix
                    + self.subrole_prefix
                    + match.group(1).replace(".", self.replace_dot)
                )
        return new_name

    def task_cb(self, task):
        """do something with a task item"""
        module_name = None
        role_module_name = None
        if "include_role" in task:
            module_name = "include_role"
        elif "import_role" in task:
            module_name = "import_role"
        elif "include_vars" in task:
            module_name = "include_vars"
        else:
            for rm in self.role_modules:
                if rm in task:
                    module_name = rm
                    role_module_name = rm
                    break
        if module_name == "include_role" or module_name == "import_role":
            new_rolename = self.convert_rolename(task[module_name]["name"])
            if new_rolename:
                task[module_name]["name"] = new_rolename
        elif module_name == "include_vars":
            """
            Convert include_vars in the test playbook.
            include_vars: path/to/{src_owner}.ROLENAME/file_or_dir
            or
            include_vars:
              file|dir: path/to/{src_owner}.ROLENAME/file_or_dir
            Note: If the path is relative and not inside a role,
            it will be parsed relative to the playbook.
            To solve it, the relative path is converted to the absolute path.
            """
            _src_owner_match = "/" + self.src_owner + "."
            _src_owner_pattern = r".*/{0}[.](\w+)/([\w\d./]+)".format(self.src_owner)
            if isinstance(task[module_name], dict):
                _key = None
                if (
                    "file" in task[module_name].keys()
                    and _src_owner_match in task[module_name]["file"]
                ):
                    _key = "file"
                elif (
                    "dir" in task[module_name].keys()
                    and _src_owner_match in task[module_name]["dir"]
                ):
                    _key = "dir"
                if _key:
                    _path = task[module_name][_key]
                    _match = re.match(_src_owner_pattern, _path)
                    task[module_name][
                        _key
                    ] = "{0}/ansible_collections/{1}/{2}/roles/{3}/{4}".format(
                        self.top_dir,
                        self.namespace,
                        self.collection,
                        _match.group(1),
                        _match.group(2),
                    )
            elif (
                isinstance(task[module_name], str)
                and _src_owner_match in task[module_name]
            ):
                _path = task[module_name]
                _match = re.match(_src_owner_pattern, _path)
                task[
                    module_name
                ] = "{0}/ansible_collections/{1}/{2}/roles/{3}/{4}".format(
                    self.top_dir,
                    self.namespace,
                    self.collection,
                    _match.group(1),
                    _match.group(2),
                )
        elif role_module_name:
            logging.debug(f"\ttask role module {role_module_name}")
            # assumes task is an orderreddict
            idx = tuple(task).index(role_module_name)
            val = task.pop(role_module_name)
            task.insert(idx, self.prefix + role_module_name, val)

    def other_cb(self, item):
        """do something with the other non-task information in an item
        this is where you will get e.g. the `roles` keyword from a play"""
        self.change_roles(item, "roles")

    def vars_cb(self, item):
        """handle vars of Ansible item, or vars from a vars file"""
        for var in item.get("vars", []):
            logging.debug(f"\tvar = {var}")
            if var == "roletoinclude":
                lsr_rolename = self.src_owner + "." + self.rolename
                if item["vars"][var] == lsr_rolename:
                    item["vars"][var] = self.prefix + self.newrolename
        return

    def meta_cb(self, item):
        """hand a meta/main.yml style file"""
        self.change_roles(item, "dependencies")

    def comp_rolenames(self, name0, name1):
        if name0 == name1:
            return True
        else:
            # self.rolename_regex is default to "[_.]".
            core0 = re.sub(self.rolename_regex, "", name0)
            core1 = re.sub(self.rolename_regex, "", name1)
            return core0 == core1

    def change_roles(self, item, roles_kw):
        """ru_item is an item which may contain a roles or dependencies
        specifier - the roles_kw is either "roles" or "dependencies"
        """
        lsr_rolename = self.src_owner + "." + self.rolename
        for idx, role in enumerate(item.get(roles_kw, [])):
            changed = False
            # role could be
            #   ordereddict([('name', 'linux-system-roles.ROLENAME')])
            # or
            #   'linux-system-roles.ROLENAME'
            if isinstance(role, dict):
                if "name" in role:
                    key = "name"
                else:
                    key = "role"
                new_rolename = self.convert_rolename(role[key], lsr_rolename)
                if new_rolename:
                    role[key] = new_rolename
                    changed = True
            else:
                new_rolename = self.convert_rolename(role, lsr_rolename)
                if new_rolename:
                    role = new_rolename
                    changed = True
            if changed:
                item[roles_kw][idx] = role

    def write(self):
        """assume we are operating on files already copied to the dest dir,
        so write file in-place"""
        self.outputfile = self.filepath
        super().write()


def lsr_copyleaf(src, dest, symlinks=True, ignore=None):
    if src.is_symlink() and symlinks:
        # symlinks=True --> symlink in dest
        copyfile(src, dest, follow_symlinks=(not symlinks))
    elif src.is_dir():
        # symlinks=True --> symlink in dest
        # symlinks=False --> copy in dest
        copytree(src, dest, symlinks=symlinks, ignore=ignore)
    else:
        copyfile(src, dest)


# Once python 3.8 is available in Travis CI,
# replace lsr_copytree with shutil.copytree with dirs_exist_ok=True.
def lsr_copytree(src, dest, symlinks=True, dirs_exist_ok=False, ignore=None):
    if dest.exists():
        if dest.is_dir():
            for sr in src.iterdir():
                subsrc = src / sr.name
                subdest = dest / sr.name
                if ignore:
                    if sr.name != ignore:
                        if subsrc.is_dir():
                            if subdest.exists() and dirs_exist_ok:
                                rmtree(subdest)
                            lsr_copytree(
                                subsrc,
                                subdest,
                                symlinks=symlinks,
                                ignore=ignore,
                                dirs_exist_ok=True,
                            )
                        else:
                            if subdest.exists() and dirs_exist_ok:
                                subdest.unlink()
                            lsr_copyleaf(
                                subsrc, subdest, symlinks=symlinks, ignore=ignore
                            )
                else:
                    if subsrc.is_dir():
                        if subdest.exists() and dirs_exist_ok:
                            rmtree(subdest)
                        lsr_copytree(
                            subsrc,
                            subdest,
                            symlinks=symlinks,
                            dirs_exist_ok=dirs_exist_ok,
                        )
                    else:
                        if (subdest.exists() or subdest.is_symlink()) and dirs_exist_ok:
                            subdest.unlink()
                        # symlinks=False --> copy in dest
                        copy2(subsrc, subdest, follow_symlinks=(not symlinks))
        else:
            if dest.exists() and dirs_exist_ok:
                dest.unlink()
            lsr_copyleaf(src, dest, symlinks=symlinks, ignore=ignore)
    else:
        lsr_copyleaf(src, dest, symlinks=symlinks, ignore=ignore)


def dir_to_plugin(v):
    if v[-8:] == "_plugins":
        return v[:-8]
    elif v == "library":
        return "modules"
    return v


def file_replace(path, find, replace, file_patterns):
    """
    Replace a pattern `find` with `replace` in the files that match
    `file_patterns` under `path`.
    """
    for root, dirs, files in os.walk(os.path.abspath(path)):
        for file_pattern in file_patterns:
            for filename in fnmatch.filter(files, file_pattern):
                filepath = os.path.join(root, filename)
                with open(filepath, encoding="utf-8") as f:
                    s = f.read()
                s = re.sub(find, replace, s)
                with open(filepath, "w", encoding="utf-8") as f:
                    f.write(s)


def copy_tree_with_replace(
    src_path,
    dest_path,
    role,
    new_role,
    TUPLE,
    transformer_args,
    isrole=True,
    ignoreme=None,
    symlinks=True,
):
    """
    1. Copy files and dirs in the dir to
       DEST_PATH/ansible_collections/NAMESPACE/COLLECTION/roles/ROLE/dir
       or
       DEST_PATH/ansible_collections/NAMESPACE/COLLECTION/dir/ROLE.
    2. Parse the source tree to look for task_roles
    3. Replace the task_roles with FQCN
    """
    for dirname in TUPLE:
        src = src_path / dirname
        if src.is_dir():
            if isrole:
                dest = dest_path / "roles" / new_role / dirname
            else:
                dest = dest_path / dirname / new_role
            logging.info(f"Copying role {src} to {dest}")
            if ignoreme:
                lsr_copytree(
                    src,
                    dest,
                    ignore=ignore_patterns(*ignoreme),
                    symlinks=symlinks,
                    dirs_exist_ok=True,
                )
            else:
                lsr_copytree(src, dest, symlinks=symlinks, dirs_exist_ok=True)
            lsrxfrm = LSRTransformer(
                dest, transformer_args, False, role, new_role, LSRFileTransformer
            )
            lsrxfrm.run()


def cleanup_symlinks(path, role, rmlist):
    """
    Clean up symlinks in tests/roles
    """
    if path.exists():
        nodes = sorted(list(path.rglob("*")), reverse=True)
        for node in nodes:
            for item in rmlist:
                if item == node.name:
                    if node.is_symlink():
                        node.unlink()
            if (
                node.is_dir()
                and r"linux-system-roles." + role == node.name
                and not any(node.iterdir())
            ):
                node.rmdir()
        roles_dir = path / "roles"
        if roles_dir.exists():
            for sr in roles_dir.iterdir():
                if sr.is_symlink():
                    sr.unlink()
            if not any(roles_dir.iterdir()):
                roles_dir.rmdir()


def gather_module_utils_parts(module_utils_dir):
    module_utils = []
    if module_utils_dir.is_dir():
        for root, dirs, files in os.walk(module_utils_dir):
            for filename in files:
                if os.path.splitext(filename)[1] != ".py":
                    continue
                full_path = (Path(root) / filename).relative_to(module_utils_dir)
                parts = bytes(full_path)[:-3].split(b"/")
                if parts[-1] == b"__init__":
                    del parts[-1]
                module_utils.append(parts)
    return module_utils


def import_replace(match):
    """
    If 'import ansible.module_utils.something ...' matches,
    'import ansible_collections.NAMESPACE.COLLECTION.plugins.module_utils.something ...'
    is returned to replace.
    """
    _src_path = config["src_path"]
    _namespace = config["namespace"]
    _collection = config["collection"]
    _role = config["role"]
    _module_utils = config["module_utils"]
    _additional_rewrites = config["additional_rewrites"]
    _module_utils_dir = config["module_utils_dir"]
    parts = match.group(3).split(b".")
    match_group3 = match.group(3)
    src_module_path = _src_path / "module_utils" / match.group(3).decode("utf-8")
    dest_module_path0 = _module_utils_dir / match.group(3).decode("utf-8")
    dest_module_path1 = _module_utils_dir / _role
    if len(parts) == 1:
        if not src_module_path.is_dir() and (
            dest_module_path0.is_dir() or dest_module_path1.is_dir()
        ):
            match_group3 = (_role + "." + match.group(3).decode("utf-8")).encode()
            parts = match_group3.split(b".")
    if parts in _module_utils:
        if match.group(1) == b"import" and match.group(4) == b"":
            _additional_rewrites.append(parts)
            if src_module_path.exists() or Path(str(src_module_path) + ".py").exists():
                return b"import ansible_collections.%s.%s.plugins.module_utils.%s%s" % (
                    bytes(_namespace, "utf-8"),
                    bytes(_collection, "utf-8"),
                    match_group3,
                    match.group(5),
                )
            else:
                return (
                    b"import ansible_collections.%s.%s.plugins.module_utils.%s as %s%s"
                    % (
                        bytes(_namespace, "utf-8"),
                        bytes(_collection, "utf-8"),
                        match_group3,
                        parts[-1],
                        match.group(5),
                    )
                )
        return b"%s ansible_collections.%s.%s.plugins.module_utils.%s%s%s" % (
            match.group(1),
            bytes(_namespace, "utf-8"),
            bytes(_collection, "utf-8"),
            match_group3,
            match.group(4),
            match.group(5),
        )
    return match.group(0)


def get_candidates(parts3, parts5):
    from_file0 = config["dest_path"] / "plugins" / "module_utils"
    for p3 in parts3:
        from_file0 = from_file0 / p3.decode("utf-8")
    from_file1 = from_file0
    for p5 in parts5:
        from_file1 = from_file1 / p5.decode("utf-8").strip(", ")
    from_file0 = Path(str(from_file0) + ".py")
    lfrom_file0 = Path(str(from_file0).lower())
    from_file1 = Path(str(from_file1) + ".py")
    lfrom_file1 = Path(str(from_file1).lower())
    return from_file0, lfrom_file0, from_file1, lfrom_file1


def from_replace(match):
    """
    case 1:
    If it matches:
      from ansible.module_utils.ROLE.somedir import module
    and if plugins/module_utils/ROLE/somedir/module.py does not exist
    in the converted tree,
    'from ansible_collections.NAMESPACE.COLLECTION.plugins.module_utils.ROLE.somedir.__init__ import module'
    is returned to replace.

    case 2:
    If it matches:
      from ansible.module_utils.ROLE.subdir.something import (\n
    and if plugins/module_utils/ROLE/subdir/something.py exists in the
    converted tree,
    'from ansible_collections.NAMESPACE.COLLECTION.plugins.module_utils.ROLE.subdir.something import (\n'
    is returned to replace.

    Legend:
    - group1 - from
    - group2 - ansible.module_utils
    - group3 - name if any
    - group4 - ( if any
    - group5 - identifier
    """
    _src_path = config["src_path"]
    _namespace = config["namespace"]
    _collection = config["collection"]
    _role = config["role"]
    _module_utils = config["module_utils"]
    _module_utils_dir = config["module_utils_dir"]
    try:
        parts3 = match.group(3).split(b".")
    except AttributeError:
        parts3 = []
    try:
        parts5 = match.group(5).split(b".")
    except AttributeError:
        parts5 = []
    # parts3 (e.g., [b'ROLE', b'subdir', b'module']) matches one module_utils or
    # size of parts3 is 1 (e.g., [b'module']), in this case, module.py was moved
    # to ROLE/module.py or module is a dir.
    # If latter, match.group(3) has to be converted to b'ROLE.module'.
    match_group3 = match.group(3)
    if len(parts3) == 1:
        src_module_path = _src_path / "module_utils" / match.group(3).decode("utf-8")
        dest_module_path0 = _module_utils_dir / match.group(3).decode("utf-8")
        dest_module_path1 = _module_utils_dir / _role
        if not src_module_path.is_dir() and (
            dest_module_path0.is_dir() or dest_module_path1.is_dir()
        ):
            match_group3 = (_role + "." + match.group(3).decode("utf-8")).encode()
            parts3 = match_group3.split(b".")
    if parts3 in _module_utils:
        from_file0, lfrom_file0, from_file1, lfrom_file1 = get_candidates(
            parts3, parts5
        )
        if (
            from_file0.is_file()
            or from_file1.is_file()
            or lfrom_file0.is_file()
            or lfrom_file1.is_file()
        ):
            return (
                b"%s ansible_collections.%s.%s.plugins.module_utils.%s import %s%s%s"
                % (
                    match.group(1),
                    bytes(_namespace, "utf-8"),
                    bytes(_collection, "utf-8"),
                    match_group3,
                    match.group(4),
                    match.group(5),
                    match.group(6),
                )
            )
        else:
            return b"%s ansible_collections.%s.%s.plugins.module_utils.%s.__init__ import %s%s%s" % (
                match.group(1),
                bytes(_namespace, "utf-8"),
                bytes(_collection, "utf-8"),
                match_group3,
                match.group(4),
                match.group(5),
                match.group(6),
            )
    if parts5 in _module_utils:
        from_file0, lfrom_file0, from_file1, lfrom_file1 = get_candidates(
            parts3, parts5
        )
        if parts3:
            if (
                from_file0.is_file()
                or from_file1.is_file()
                or lfrom_file0.is_file()
                or lfrom_file1.is_file()
            ):
                return b"%s ansible_collections.%s.%s.plugins.module_utils.%s import %s%s%s" % (
                    match.group(1),
                    bytes(_namespace, "utf-8"),
                    bytes(_collection, "utf-8"),
                    match.group(3),
                    match.group(4),
                    match.group(5),
                    match.group(6),
                )
            else:
                return b"%s ansible_collections.%s.%s.plugins.module_utils.%s.__init__ import %s%s%s" % (
                    match.group(1),
                    bytes(_namespace, "utf-8"),
                    bytes(_collection, "utf-8"),
                    match.group(3),
                    match.group(4),
                    match.group(5),
                    match.group(6),
                )
        if (
            from_file0.is_file()
            or from_file1.is_file()
            or lfrom_file0.is_file()
            or lfrom_file1.is_file()
        ):
            return (
                b"%s ansible_collections.%s.%s.plugins.module_utils import %s%s%s"
                % (
                    match.group(1),
                    bytes(_namespace, "utf-8"),
                    bytes(_collection, "utf-8"),
                    match.group(4),
                    match.group(5),
                    match.group(6),
                )
            )
        else:
            return b"%s ansible_collections.%s.%s.plugins.module_utils.__init__ import %s%s%s" % (
                match.group(1),
                bytes(_namespace, "utf-8"),
                bytes(_collection, "utf-8"),
                match.group(4),
                match.group(5),
                match.group(6),
            )
    return match.group(0)


def add_rolename(filename, rolename):
    """
    A file with an extension, e.g., README.md is converted to README-rolename.md
    A file with no extension, e.g., LICENSE is to LICENSE-rolename
    """
    if filename.find(".", 1) > 0:
        with_rolename = re.sub(
            r"([\w\d_\.]+)(\.)([\w\d]*)",
            r"\1" + "-" + rolename + r"\2" + r"\3",
            filename,
        )
    else:
        with_rolename = filename + "-" + rolename
    return with_rolename


config = {}


def role2collection():
    HOME = os.environ.get("HOME")
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--namespace",
        type=str,
        default=os.environ.get("COLLECTION_NAMESPACE", "fedora"),
        help="Collection namespace; default to fedora",
    )
    parser.add_argument(
        "--collection",
        type=str,
        default=os.environ.get("COLLECTION_NAME", "linux_system_roles"),
        help="Collection name; default to linux_system_roles",
    )
    parser.add_argument(
        "--dest-path",
        type=Path,
        default=os.environ.get("COLLECTION_DEST_PATH", HOME + "/.ansible/collections"),
        help="Path to parent of collection where role should be migrated; default to ${HOME}/.ansible/collections",
    )
    parser.add_argument(
        "--tests-dest-path",
        type=Path,
        default=os.environ.get("COLLECTION_TESTS_DEST_PATH", None),
        help="Path to parent of tests directory in which rolename directory is created and test scripts are copied to the directory; default to DEST_PATH/NAMESPACE/COLLECTION",  # noqa:E501
    )
    parser.add_argument(
        "--src-path",
        type=Path,
        default=os.environ.get("COLLECTION_SRC_PATH", HOME + "/linux-system-roles"),
        help="Path to the parent directory of the source role; default to ${HOME}/linux-system-roles",
    )
    parser.add_argument(
        "--src-owner",
        type=str,
        default=os.environ.get("COLLECTION_SRC_OWNER", ""),
        help='Owner of the role in github. If the parent directory name in SRC_PATH is not the github owner, may need to set to it, e.g., "linux-system-roles"; default to the parent directory of SRC_PATH',  # noqa:E501
    )
    parser.add_argument(
        "--role",
        type=str,
        default=os.environ.get("COLLECTION_ROLE"),
        help="Role to convert to collection",
    )
    parser.add_argument(
        "--new-role",
        type=str,
        default=os.environ.get("COLLECTION_NEW_ROLE"),
        help="Role to convert to collection; specify if different from the original role; default to the value of '--role'.",  # noqa:E501
    )
    parser.add_argument(
        "--replace-dot",
        type=str,
        default=os.environ.get("COLLECTION_REPLACE_DOT", "_"),
        help=(
            "If sub-role name contains dots, replace them with the specified value; "
            "default to '_'"
        ),
    )
    parser.add_argument(
        "--subrole-prefix",
        type=str,
        default=os.environ.get("COLLECTION_SUBROLE_PREFIX", ""),
        help=(
            "If sub-role name does not start with the specified value, "
            "change the name to start with the value; default to an empty string"
        ),
    )
    parser.add_argument(
        "--readme",
        type=Path,
        default=os.environ.get("COLLECTION_README", None),
        help="Path to the readme file used in top README.md",
    )
    parser.add_argument(
        "--extra-mapping",
        type=str,
        default=os.environ.get("COLLECTION_EXTRA_MAPPING", ""),
        help=(
            "This is a comma delimited list of extra mappings to apply when "
            "converting the files - this converts the given name to collection "
            "format with the optional given namespace and collection."
        ),
    )
    args, unknown = parser.parse_known_args()

    role = args.role
    if not role:
        parser.print_help()
        logging.error("Message: role is not specified.")
        os._exit(errno.EINVAL)

    new_role = args.new_role
    if not new_role:
        new_role = role

    namespace = args.namespace
    collection = args.collection
    prefix = namespace + "." + collection + "."
    top_dest_path = args.dest_path.resolve()
    current_dest = os.path.expanduser(str(top_dest_path))
    replace_dot = args.replace_dot
    subrole_prefix = args.subrole_prefix
    readme_path = args.readme

    # Input: "linux-system-roles.role0:newrole0,\
    #         linux-system-roles.role1:fedora.linux_system_roles.role1"
    # Output:
    # [
    #   {'src_name': {'src_owner': 'linux-system-roles', 'role': 'sap-base-settings'},
    #    'dest_name': {'dest_prefix': None, 'role': ['base_settings']}},
    #   {'src_name': {'src_owner': 'linux-system-roles', 'role': 'timesync'},
    #    'dest_name': {'dest_prefix': 'fedora.linux_system_roles', 'role': 'timesync'}}
    # ]
    # Note: skip if src role is the role to be converted.
    def parse_extra_mapping(mapping_str, namespace, role):
        _mapping_list = mapping_str.split(",")
        _mapping_dict_list = []
        for _map in _mapping_list:
            _item = _map.split(":")
            if len(_item) == 2:
                _src = _item[0].split(".")
                if len(_src) == 1 or len(_src) == 2:
                    _src_name = {}
                    if len(_src) == 1:
                        # "rolename"
                        _src_name["src_owner"] = None
                        _src_name["role"] = _src[0]
                    elif len(_src) == 2:
                        # "linux-system-roles.rolename"
                        _src_name["src_owner"] = _src[0]
                        _src_name["role"] = _src[1]
                    _dest = _item[1].split(".")
                    if len(_dest) > 0 and len(_dest) <= 3:
                        _dest_name = {}
                        if len(_dest) == 1:
                            # "rolename"
                            _dest_name["dest_prefix"] = None
                            _dest_name["role"] = _dest[0]
                        elif len(_dest) == 2:
                            # "collection.rolename"
                            _dest_name["dest_prefix"] = "{0}.{1}.".format(
                                namespace, _dest[0]
                            )
                            _dest_name["role"] = _dest[1]
                        elif len(_dest) == 3:
                            # "namespace.collection.rolename"
                            _dest_name["dest_prefix"] = "{0}.{1}.".format(
                                _dest[0], _dest[1]
                            )
                            _dest_name["role"] = _dest[2]

                        _mapping_dict = {}
                        _mapping_dict["src_name"] = _src_name
                        _mapping_dict["dest_name"] = _dest_name
                        _mapping_dict_list.append(_mapping_dict)
        return _mapping_dict_list

    extra_mapping = parse_extra_mapping(args.extra_mapping, namespace, role)
    extra_mapping_src_owner = list(
        map(itemgetter("src_owner"), list(map(itemgetter("src_name"), extra_mapping)))
    )
    extra_mapping_src_role = list(
        map(itemgetter("role"), list(map(itemgetter("src_name"), extra_mapping)))
    )
    extra_mapping_dest_prefix = list(
        map(
            itemgetter("dest_prefix"), list(map(itemgetter("dest_name"), extra_mapping))
        )
    )
    extra_mapping_dest_role = list(
        map(itemgetter("role"), list(map(itemgetter("dest_name"), extra_mapping)))
    )

    dest_path = Path.joinpath(
        top_dest_path, "ansible_collections/" + namespace + "/" + collection
    )
    _tests_dest_path = args.tests_dest_path
    if _tests_dest_path:
        tests_dest_path = Path(_tests_dest_path)
    else:
        tests_dest_path = dest_path

    os.makedirs(dest_path, exist_ok=True)

    roles_dir = dest_path / "roles"
    tests_dir = tests_dest_path / "tests"
    plugin_dir = dest_path / "plugins"
    modules_dir = plugin_dir / "modules"
    module_utils_dir = plugin_dir / "module_utils"
    docs_dir = dest_path / "docs"

    src_path = args.src_path.resolve()
    src_owner = args.src_owner
    if not src_owner:
        src_owner = os.path.basename(src_path)
    _tasks_main = src_path / "tasks/main.yml"
    if not _tasks_main.exists():
        src_path = src_path / role
        _tasks_main = src_path / "tasks/main.yml"

    if not _tasks_main.exists():
        logging.error(
            f"Neither {src_path} nor {src_path.parent} is a role top directory."
        )
        sys.exit(errno.ENOENT)
    _extras = set(os.listdir(src_path)).difference(ALL_DIRS)
    try:
        _extras.remove(".git")
    except KeyError:
        pass
    extras = [src_path / e for e in _extras]

    global config
    config = {
        "namespace": namespace,
        "collection": collection,
        "role": new_role,
        "src_path": src_path,
        "dest_path": dest_path,
        "module_utils_dir": module_utils_dir,
    }

    transformer_args = {
        "namespace": namespace,
        "collection": collection,
        "prefix": prefix,
        "subrole_prefix": subrole_prefix,
        "replace_dot": replace_dot,
        # get role modules - will need to find and convert these to use FQCN
        "role_modules": get_role_modules(src_path),
        "src_owner": src_owner,
        "top_dir": current_dest,
        "extra_mapping_src_owner": extra_mapping_src_owner,
        "extra_mapping_src_role": extra_mapping_src_role,
        "extra_mapping_dest_prefix": extra_mapping_dest_prefix,
        "extra_mapping_dest_role": extra_mapping_dest_role,
    }

    # Role - copy subdirectories, tasks, defaults, vars, etc., in the system role to
    # DEST_PATH/ansible_collections/NAMESPACE/COLLECTION/roles/ROLE.
    copy_tree_with_replace(
        src_path, dest_path, role, new_role, ROLE_DIRS, transformer_args
    )

    # ==============================================================================

    copy_tree_with_replace(
        src_path,
        tests_dest_path,
        role,
        new_role,
        TESTS,
        transformer_args,
        isrole=False,
        ignoreme=["artifacts", "linux-system-roles.*", "__pycache__", ".git*"],
    )

    # remove symlinks in the tests/role.
    removeme = ["library", "modules", "module_utils", "roles"]
    cleanup_symlinks(tests_dir / new_role, role, removeme)

    # ==============================================================================

    def update_readme(src_path, filename, rolename, comment, issubrole=False):
        if not filename.startswith("README"):
            return
        if filename == "README.md":
            title = rolename
        else:
            m = re.match(r"README(.*)(\.md)", filename)
            title = rolename + m.group(1)
        main_doc = dest_path / "README.md"
        if not main_doc.exists():
            if readme_path and Path(readme_path).exists():
                with open(readme_path, encoding="utf-8") as f:
                    _s = f.read()
            else:
                _s = textwrap.dedent(
                    """\
                    # {0} {1} collections
                    """
                ).format(namespace, collection)

            s = textwrap.dedent(
                """\
                {0}

                {1}

                <!--ts-->
                  * {2}
                <!--te-->
                """
            ).format(_s, comment, title)
            with open(main_doc, "w", encoding="utf-8") as f:
                f.write(s)
        else:
            with open(main_doc, encoding="utf-8") as f:
                s = f.read()
            role_link = "{0}".format(title)
            if role_link not in s:
                if comment not in s:
                    text = (
                        s
                        + textwrap.dedent(
                            """\

                        {2}

                        <!--ts-->
                          * {3}
                        <!--te-->
                        """
                        ).format(namespace, collection, comment, role_link)
                    )
                else:
                    find = r"({0}\n\n<!--ts-->\n)(( |\*|\w|\[|\]|\(|\)|\.|/|-|\n|\r)+)".format(
                        comment
                    )
                    replace = r"\1\2  * {0}\n".format(role_link)
                    text = re.sub(find, replace, s, flags=re.M)
                with open(main_doc, "w", encoding="utf-8") as f:
                    f.write(text)

    # Copy docs, design_docs, and examples to
    # DEST_PATH/ansible_collections/NAMESPACE/COLLECTION/docs/ROLE.
    # Copy README.md to DEST_PATH/ansible_collections/NAMESPACE/COLLECTION/roles/ROLE.
    # Generate a top level README.md which contains links to roles/ROLE/README.md.
    def process_readme(
        src_path, filename, role, new_role, original=None, issubrole=False
    ):
        """
        Copy src_path/filename to dest_path/docs/new_role.
        filename could be README.md, README-something.md, or something.md.
        Create a primary README.md in dest_path, which points to dest_path/docs/new_role/filename
        with the title new_role or new_role-something.
        """
        src = src_path / filename
        dest = roles_dir / new_role / filename
        # copy
        logging.info(f"Copying doc {filename} to {dest}")
        copy2(src, dest, follow_symlinks=False)
        dest = roles_dir / new_role
        file_patterns = ["*.md"]
        file_replace(dest, src_owner + "." + role, prefix + new_role, file_patterns)
        # --extra-mapping SRCROLENAME:DESTROLENAME(or FQCN)
        for _emap in extra_mapping:
            # Replacing SRC_OWNER.ROLE with FQCN
            _from = "{0}.{1}".format(
                _emap["src_name"]["src_owner"]
                if _emap["src_name"]["src_owner"]
                else src_owner,
                _emap["src_name"]["role"],
            )
            _to = "{0}{1}".format(
                _emap["dest_name"]["dest_prefix"]
                if _emap["dest_name"]["dest_prefix"]
                else prefix,
                _emap["dest_name"]["role"],
            )
            file_replace(dest, _from, _to, file_patterns)
            # Replacing unprefixed ROLE with FQCN
            _from = " {0}".format(_emap["src_name"]["role"])
            _to = " {0}{1}".format(
                _emap["dest_name"]["dest_prefix"]
                if _emap["dest_name"]["dest_prefix"]
                else prefix,
                _emap["dest_name"]["role"],
            )
            file_replace(dest, _from, _to, file_patterns)
        if original:
            file_replace(dest, original, prefix + new_role, file_patterns)
        if filename == "README.md":
            if issubrole:
                comment = "### Private Roles"
            else:
                comment = "### Supported Roles"
            update_readme(src_path, filename, new_role, comment, issubrole)

    ignoreme = ["linux-system-roles.*"]
    dest = docs_dir / new_role
    for doc in DOCS:
        src = src_path / doc
        if src.is_dir():
            logging.info(f"Copying docs {src} to {dest}")
            lsr_copytree(
                src,
                dest,
                symlinks=False,
                ignore=ignore_patterns(*ignoreme),
                dirs_exist_ok=True,
            )
            if doc == "examples":
                lsrxfrm = LSRTransformer(
                    dest,
                    transformer_args,
                    False,
                    role,
                    new_role,
                    LSRFileTransformer,
                )
                lsrxfrm.run()
        elif src.is_file():
            process_readme(src_path, doc, role, new_role)

    # Remove symlinks in the docs/role (e.g., in the examples).
    removeme = ["library", "modules", "module_utils", "roles"]
    cleanup_symlinks(dest, new_role, removeme)

    # ==============================================================================

    # Copy library, module_utils, plugins
    # Library and plugins are copied to dest_path/plugins
    # If plugin is in SUBDIR (currently, just module_utils),
    #   module_utils/*.py are to dest_path/plugins/module_utils/ROLE/*.py
    #   module_utils/subdir/*.py are to dest_path/plugins/module_utils/subdir/*.py
    SUBDIR = ("module_utils",)
    for plugin in PLUGINS:
        src = src_path / plugin
        plugin_name = dir_to_plugin(plugin)
        if not src.is_dir():
            continue
        if plugin in SUBDIR:
            for sr in src.iterdir():
                if sr.is_dir():
                    # If src/sr is a directory, copy it to the dest
                    dest = plugin_dir / plugin_name / sr.name
                    logging.info(f"Copying plugin {sr} to {dest}")
                    lsr_copytree(sr, dest)
                else:
                    # Otherwise, copy it to the plugins/plugin_name/ROLE
                    dest = plugin_dir / plugin_name / new_role
                    dest.mkdir(parents=True, exist_ok=True)
                    logging.info(f"Copying plugin {sr} to {dest}")
                    copy2(sr, dest, follow_symlinks=False)
        else:
            dest = plugin_dir / plugin_name
            logging.info(f"Copying plugin {src} to {dest}")
            lsr_copytree(src, dest)

    # Update the python codes which import modules in plugins/{modules,modules_dir}.
    config["module_utils"] = gather_module_utils_parts(module_utils_dir)
    additional_rewrites = []
    config["additional_rewrites"] = additional_rewrites
    for rewrite_dir in (module_utils_dir, modules_dir):
        if rewrite_dir.is_dir():
            for root, dirs, files in os.walk(rewrite_dir):
                for filename in files:
                    if os.path.splitext(filename)[1] != ".py":
                        continue
                    full_path = Path(root) / filename
                    text = full_path.read_bytes()
                    new_text = IMPORT_RE.sub(import_replace, text)
                    new_text = FROM_RE.sub(from_replace, new_text)
                    for rewrite in additional_rewrites:
                        pattern = re.compile(
                            re.escape(br"ansible.module_utils.%s" % b".".join(rewrite))
                        )
                        new_text = pattern.sub(rewrite[-1], new_text)

                    if text != new_text:
                        logging.info("Rewriting imports for {}".format(full_path))
                        full_path.write_bytes(new_text)
                        additional_rewrites[:] = []

    # ==============================================================================

    # Extra files and directories including the sub-roles
    for extra in extras:
        if extra.name in TOX:
            continue
        if extra.name.endswith(".md"):
            # E.g., contributing.md, README-devel.md and README-testing.md
            process_readme(extra.parent, extra.name, role, new_role)
        elif extra.is_dir():
            # Copying sub-roles to the roles dir and its tests and README are also
            # handled in the same way as the parent role's are.
            if extra.name == "roles":
                for sr in extra.iterdir():
                    # If a role name contains '.', replace it with replace_dot
                    # convert nested subroles to prefix name with subrole_prefix.
                    dr = sr.name.replace(".", replace_dot)
                    if subrole_prefix and not dr.startswith(subrole_prefix):
                        dr = subrole_prefix + dr
                    copy_tree_with_replace(
                        sr, dest_path, dr, dr, ROLE_DIRS, transformer_args
                    )
                    # copy tests dir to dest_path/"tests"
                    copy_tree_with_replace(
                        sr,
                        tests_dest_path,
                        dr,
                        dr,
                        TESTS,
                        transformer_args,
                        isrole=False,
                        ignoreme=[
                            "artifacts",
                            "linux-system-roles.*",
                            "__pycache__",
                            ".git*",
                        ],
                    )
                    # remove symlinks in the tests/new_role.
                    removeme = ["library", "modules", "module_utils", "roles"]
                    cleanup_symlinks(tests_dir / dr, dr, removeme)
                    # copy README.md to dest_path/roles/sr.name
                    _readme = sr / "README.md"
                    if _readme.is_file():
                        process_readme(
                            sr, "README.md", dr, dr, original=sr.name, issubrole=True
                        )
                    if sr.name != dr:
                        # replace "sr.name" with "dr" in role_dir
                        dirs = ["roles", "docs", "tests"]
                        for dir in dirs:
                            role_dir = dest_path / dir
                            file_patterns = ["*.yml", "*.md"]
                            file_replace(
                                role_dir,
                                re.escape("\b" + sr.name + "\b"),
                                dr,
                                file_patterns,
                            )
            # Other extra directories are copied to the collection dir as they are.
            else:
                dest = dest_path / extra.name
                logging.info(f"Copying extra {extra} to {dest}")
                lsr_copytree(extra, dest)
        # Other extra files.
        else:
            if extra.name.endswith(".yml") and "playbook" in extra.name:
                # some-playbook.yml is copied to docs/role dir.
                dest = dest_path / "docs" / new_role
                dest.mkdir(parents=True, exist_ok=True)
            else:
                # If the extra file 'filename' has no extension, it is copied to the collection dir as
                # 'filename-ROLE'. If the extra file is 'filename.ext', it is copied to 'filename-ROLE.ext'.
                dest = dest_path / add_rolename(extra.name, new_role)
            logging.info(f"Copying extra {extra} to {dest}")
            copy2(extra, dest, follow_symlinks=False)

    dest = dest_path / "docs" / new_role
    if dest.is_dir():
        lsrxfrm = LSRTransformer(
            dest, transformer_args, False, role, new_role, LSRFileTransformer
        )
        lsrxfrm.run()

    default_collections_paths = "~/.ansible/collections:/usr/share/ansible/collections"
    default_collections_paths_list = list(
        map(os.path.expanduser, default_collections_paths.split(":"))
    )
    # top_dest_path is not in the default collections path.
    # suggest to run ansible-playbook with ANSIBLE_COLLECTIONS_PATHS env var.
    if current_dest not in default_collections_paths_list:
        ansible_collections_paths = current_dest + ":" + default_collections_paths
        print(
            f"Run ansible-playbook with environment variable ANSIBLE_COLLECTIONS_PATHS={ansible_collections_paths}"
        )
    return 0


if __name__ == "__main__":
    sys.exit(role2collection())



# SPDX-License-Identifier: GPL-2.0-or-later
"""unit tests for lsr_role2collection"""

import os
import re
import shutil
import tempfile
import textwrap
from pathlib import Path
import unittest

from lsr_role2collection import (
    file_replace,
    copy_tree_with_replace,
    import_replace,
    from_replace,
    gather_module_utils_parts,
    add_rolename,
    config,
)

src_path = os.environ.get("COLLECTION_SRC_PATH", "/var/tmp/linux-system-roles")
dest_path = os.environ.get("COLLECTION_DEST_PATH", "/var/tmp/collections")
namespace = os.environ.get("COLLECTION_NAMESPACE", "fedora")
collection_name = os.environ.get("COLLECTION_NAME", "system_roles")
rolename = "systemrole"
newrolename = "newsystemrole"
otherroles = ["other0", "other1", "other2"]
newotherroles = ["newother0", "newother1", "newother2"]
prefix = namespace + "." + collection_name
prefixdot = prefix + "."
otherprefix = "mynamespace.mycollection"
otherprefixdot = "mynamespace.mycollection."

test_yaml_str = textwrap.dedent(
    """\
    # SPDX-License-Identifier: MIT
    ---
    - name: Ensure that the role runs with default parameters
      hosts: all
      {0}:
        - {1}
            {2} {3}{4}{5}
    """
)


class LSRRole2Collection(unittest.TestCase):
    """test lsr_role2collection"""

    def create_test_tree(self, top, template, params, ext, is_vertical=True):
        path = top
        for n in range(len(params)):
            if is_vertical:
                path /= "sub" + str(n)
            path.mkdir(parents=True, exist_ok=True)
            filename = "test" + str(n) + ext
            filepath = path / filename
            content = template.format(
                params[n]["keyword"],
                params[n]["role_or_task_name"],
                params[n]["task_subkey"],
                params[n]["task_value"],
                params[n]["task_delim"],
                params[n]["task_subvalue"],
            )
            with open(filepath, "w") as f:
                f.write(content)

    def check_test_tree(self, top, template, params, ext, is_vertical=True):
        path = top
        for n in range(len(params)):
            if is_vertical:
                path /= "sub" + str(n)
            filename = "test" + str(n) + ext
            filepath = path / filename
            content = template.format(
                params[n]["keyword"],
                params[n]["role_or_task_name"],
                params[n]["task_subkey"],
                params[n]["task_value"],
                params[n]["task_delim"],
                params[n]["task_subvalue"],
            )
            with open(filepath) as f:
                s = f.read()
            self.assertEqual(content.rstrip(), s.rstrip())

    def create_test_link(self, path, link, target, is_target):
        path.mkdir(parents=True, exist_ok=True)
        (path / link).symlink_to(target, target_is_directory=is_target)

    def check_test_link(self, path, exists):
        self.assertEqual(path.exists(), exists)

    def test_file_replace(self):
        """test file_replace"""

        tmpdir = tempfile.TemporaryDirectory()
        pre_params = [
            {
                "keyword": "roles",
                "role_or_task_name": "linux-system-roles." + rolename,
                "task_subkey": "",
                "task_value": "",
                "task_delim": "",
                "task_subvalue": "",
            },
            {
                "keyword": "tasks",
                "role_or_task_name": "import_role:",
                "task_subkey": "name:",
                "task_value": "linux-system-roles",
                "task_delim": ".",
                "task_subvalue": rolename,
            },
            {
                "keyword": "tasks",
                "role_or_task_name": "include_role:",
                "task_subkey": "name:",
                "task_value": "linux-system-roles",
                "task_delim": ".",
                "task_subvalue": rolename,
            },
        ]
        post_params = [
            {
                "keyword": "roles",
                "role_or_task_name": prefix + "." + rolename,
                "task_subkey": "",
                "task_value": "",
                "task_delim": "",
                "task_subvalue": "",
            },
            {
                "keyword": "tasks",
                "role_or_task_name": "import_role:",
                "task_subkey": "name:",
                "task_value": prefix,
                "task_delim": ".",
                "task_subvalue": rolename,
            },
            {
                "keyword": "tasks",
                "role_or_task_name": "include_role:",
                "task_subkey": "name:",
                "task_value": prefix,
                "task_delim": ".",
                "task_subvalue": rolename,
            },
        ]
        self.create_test_tree(Path(tmpdir.name), test_yaml_str, pre_params, ".yml")
        file_replace(
            Path(tmpdir.name),
            pre_params[1]["task_value"] + "." + pre_params[1]["task_subvalue"],
            prefixdot + pre_params[1]["task_subvalue"],
            ["*.yml"],
        )
        self.check_test_tree(Path(tmpdir.name), test_yaml_str, post_params, ".yml")

    def test_copy_tree_with_replace(self):
        """test copy_tree_with_replace"""

        pre_params = [
            {
                "keyword": "roles",
                "role_or_task_name": "linux-system-roles." + rolename,
                "task_subkey": "",
                "task_value": "",
                "task_delim": "",
                "task_subvalue": "",
            },
            {
                "keyword": "tasks",
                "role_or_task_name": "import_role:",
                "task_subkey": "name:",
                "task_value": "linux-system-roles",
                "task_delim": ".",
                "task_subvalue": rolename,
            },
            {
                "keyword": "tasks",
                "role_or_task_name": "include_role:",
                "task_subkey": "name:",
                "task_value": "linux-system-roles",
                "task_delim": ".",
                "task_subvalue": rolename,
            },
            {
                "keyword": "tasks",
                "role_or_task_name": "include_role:",
                "task_subkey": "name:",
                "task_value": '"{{ role_path }}/roles',
                "task_delim": "/",
                "task_subvalue": '__subrole_with__"',
            },
            {
                "keyword": "tasks",
                "role_or_task_name": "include_role:",
                "task_subkey": "name:",
                "task_value": '"{{ role_path }}/roles',
                "task_delim": "/",
                "task_subvalue": 'subrole_no__"',
            },
            {
                "keyword": "tasks",
                "role_or_task_name": "include_tasks:",
                "task_subkey": "-",
                "task_value": '"{{ role_path }}/roles/subrole/tasks',
                "task_delim": "/",
                "task_subvalue": 'mytask.yml"',
            },
            {
                "keyword": "tasks",
                "role_or_task_name": "include_vars:",
                "task_subkey": "-",
                "task_value": '"{{ role_path }}/roles/subrole/vars',
                "task_delim": "/",
                "task_subvalue": 'myvar.yml"',
            },
        ]
        post_params = [
            {
                "keyword": "roles",
                "role_or_task_name": prefix + "." + rolename,
                "task_subkey": "",
                "task_value": "",
                "task_delim": "",
                "task_subvalue": "",
            },
            {
                "keyword": "tasks",
                "role_or_task_name": "import_role:",
                "task_subkey": "name:",
                "task_value": prefix,
                "task_delim": ".",
                "task_subvalue": rolename,
            },
            {
                "keyword": "tasks",
                "role_or_task_name": "include_role:",
                "task_subkey": "name:",
                "task_value": prefix,
                "task_delim": ".",
                "task_subvalue": rolename,
            },
            {
                "keyword": "tasks",
                "role_or_task_name": "include_role:",
                "task_subkey": "name:",
                "task_value": '"' + prefix,
                "task_delim": ".",
                "task_subvalue": '__subrole_with__"',
            },
            {
                "keyword": "tasks",
                "role_or_task_name": "include_role:",
                "task_subkey": "name:",
                "task_value": '"' + prefix,
                "task_delim": ".",
                "task_subvalue": 'subrole_no__"',
            },
            {
                "keyword": "tasks",
                "role_or_task_name": "include_tasks:",
                "task_subkey": "-",
                "task_value": '"{{ role_path }}/roles/subrole/tasks',
                "task_delim": "/",
                "task_subvalue": 'mytask.yml"',
            },
            {
                "keyword": "tasks",
                "role_or_task_name": "include_vars:",
                "task_subkey": "-",
                "task_value": '"{{ role_path }}/roles/subrole/vars',
                "task_delim": "/",
                "task_subvalue": 'myvar.yml"',
            },
        ]
        MYTUPLE = ("tasks",)
        tmpdir = tempfile.TemporaryDirectory()
        role_path = Path(tmpdir.name) / "linux-system-roles" / rolename
        coll_path = (
            Path(dest_path) / "ansible_collections" / namespace / collection_name
        )
        self.create_test_tree(
            role_path / "tasks", test_yaml_str, pre_params, ".yml", is_vertical=False
        )

        transformer_args = {
            "namespace": namespace,
            "collection": collection_name,
            "prefix": prefixdot,
            "subrole_prefix": "",
            "replace_dot": "_",
            "role_modules": set(),
            "src_owner": "linux-system-roles",
            "top_dir": dest_path,
            "extra_mapping_src_owner": [],
            "extra_mapping_src_role": [],
            "extra_mapping_dest_prefix": [],
            "extra_mapping_dest_role": [],
        }
        copy_tree_with_replace(
            role_path,
            coll_path,
            rolename,
            rolename,
            MYTUPLE,
            transformer_args,
            isrole=True,
        )
        test_path = coll_path / "roles" / rolename / "tasks"
        self.check_test_tree(
            test_path, test_yaml_str, post_params, ".yml", is_vertical=False
        )
        shutil.rmtree(coll_path)

    def test_copy_tree_with_replace_with_newrolename(self):
        """test copy_tree_with_replace"""

        pre_params = [
            {
                "keyword": "roles",
                "role_or_task_name": "linux-system-roles." + rolename,
                "task_subkey": "",
                "task_value": "",
                "task_delim": "",
                "task_subvalue": "",
            },
            {
                "keyword": "tasks",
                "role_or_task_name": "import_role:",
                "task_subkey": "name:",
                "task_value": "linux-system-roles",
                "task_delim": ".",
                "task_subvalue": rolename,
            },
            {
                "keyword": "tasks",
                "role_or_task_name": "include_role:",
                "task_subkey": "name:",
                "task_value": "linux-system-roles",
                "task_delim": ".",
                "task_subvalue": rolename,
            },
            {
                "keyword": "tasks",
                "role_or_task_name": "include_role:",
                "task_subkey": "name:",
                "task_value": '"{{ role_path }}/roles',
                "task_delim": "/",
                "task_subvalue": '__subrole_with__"',
            },
            {
                "keyword": "tasks",
                "role_or_task_name": "include_role:",
                "task_subkey": "name:",
                "task_value": '"{{ role_path }}/roles',
                "task_delim": "/",
                "task_subvalue": 'subrole_no__"',
            },
            {
                "keyword": "tasks",
                "role_or_task_name": "include_tasks:",
                "task_subkey": "-",
                "task_value": '"{{ role_path }}/roles/subrole/tasks',
                "task_delim": "/",
                "task_subvalue": 'mytask.yml"',
            },
            {
                "keyword": "tasks",
                "role_or_task_name": "include_vars:",
                "task_subkey": "-",
                "task_value": '"{{ role_path }}/roles/subrole/vars',
                "task_delim": "/",
                "task_subvalue": 'myvar.yml"',
            },
        ]
        post_params = [
            {
                "keyword": "roles",
                "role_or_task_name": prefix + "." + newrolename,
                "task_subkey": "",
                "task_value": "",
                "task_delim": "",
                "task_subvalue": "",
            },
            {
                "keyword": "tasks",
                "role_or_task_name": "import_role:",
                "task_subkey": "name:",
                "task_value": prefix,
                "task_delim": ".",
                "task_subvalue": newrolename,
            },
            {
                "keyword": "tasks",
                "role_or_task_name": "include_role:",
                "task_subkey": "name:",
                "task_value": prefix,
                "task_delim": ".",
                "task_subvalue": newrolename,
            },
            {
                "keyword": "tasks",
                "role_or_task_name": "include_role:",
                "task_subkey": "name:",
                "task_value": '"' + prefix,
                "task_delim": ".",
                "task_subvalue": '__subrole_with__"',
            },
            {
                "keyword": "tasks",
                "role_or_task_name": "include_role:",
                "task_subkey": "name:",
                "task_value": '"' + prefix,
                "task_delim": ".",
                "task_subvalue": 'subrole_no__"',
            },
            {
                "keyword": "tasks",
                "role_or_task_name": "include_tasks:",
                "task_subkey": "-",
                "task_value": '"{{ role_path }}/roles/subrole/tasks',
                "task_delim": "/",
                "task_subvalue": 'mytask.yml"',
            },
            {
                "keyword": "tasks",
                "role_or_task_name": "include_vars:",
                "task_subkey": "-",
                "task_value": '"{{ role_path }}/roles/subrole/vars',
                "task_delim": "/",
                "task_subvalue": 'myvar.yml"',
            },
        ]
        MYTUPLE = ("tasks",)
        tmpdir = tempfile.TemporaryDirectory()
        role_path = Path(tmpdir.name) / "linux-system-roles" / rolename
        coll_path = (
            Path(dest_path) / "ansible_collections" / namespace / collection_name
        )
        self.create_test_tree(
            role_path / "tasks", test_yaml_str, pre_params, ".yml", is_vertical=False
        )
        transformer_args = {
            "namespace": namespace,
            "collection": collection_name,
            "prefix": prefixdot,
            "subrole_prefix": "",
            "replace_dot": "_",
            "role_modules": set(),
            "src_owner": "linux-system-roles",
            "top_dir": dest_path,
            "extra_mapping_src_owner": [],
            "extra_mapping_src_role": [],
            "extra_mapping_dest_prefix": [],
            "extra_mapping_dest_role": [],
        }
        copy_tree_with_replace(
            role_path,
            coll_path,
            rolename,
            newrolename,
            MYTUPLE,
            transformer_args,
            isrole=True,
        )
        test_path = coll_path / "roles" / newrolename / "tasks"
        self.check_test_tree(
            test_path, test_yaml_str, post_params, ".yml", is_vertical=False
        )
        shutil.rmtree(coll_path)

    def test_copy_tree_with_replace_extra_mapping(self):
        """test copy_tree_with_replace with extra_mapping"""

        pre_params = [
            {
                "keyword": "roles",
                "role_or_task_name": "linux-system-roles." + otherroles[0],
                "task_subkey": "",
                "task_value": "",
                "task_delim": "",
                "task_subvalue": "",
            },
            {
                "keyword": "tasks",
                "role_or_task_name": "import_role:",
                "task_subkey": "name:",
                "task_value": "linux-system-roles",
                "task_delim": ".",
                "task_subvalue": otherroles[1],
            },
            {
                "keyword": "tasks",
                "role_or_task_name": "include_role:",
                "task_subkey": "name:",
                "task_value": "linux-system-roles",
                "task_delim": ".",
                "task_subvalue": otherroles[2],
            },
        ]
        post_params = [
            {
                "keyword": "roles",
                "role_or_task_name": otherprefix + "." + newotherroles[0],
                "task_subkey": "",
                "task_value": "",
                "task_delim": "",
                "task_subvalue": "",
            },
            {
                "keyword": "tasks",
                "role_or_task_name": "import_role:",
                "task_subkey": "name:",
                "task_value": otherprefix,
                "task_delim": ".",
                "task_subvalue": newotherroles[1],
            },
            {
                "keyword": "tasks",
                "role_or_task_name": "include_role:",
                "task_subkey": "name:",
                "task_value": prefix,
                "task_delim": ".",
                "task_subvalue": newotherroles[2],
            },
        ]
        MYTUPLE = ("tasks",)
        tmpdir = tempfile.TemporaryDirectory()
        role_path = Path(tmpdir.name) / "linux-system-roles" / rolename
        coll_path = (
            Path(dest_path) / "ansible_collections" / namespace / collection_name
        )
        self.create_test_tree(
            role_path / "tasks", test_yaml_str, pre_params, ".yml", is_vertical=False
        )
        transformer_args = {
            "namespace": namespace,
            "collection": collection_name,
            "prefix": prefixdot,
            "subrole_prefix": "",
            "replace_dot": "_",
            "role_modules": set(),
            "src_owner": "linux-system-roles",
            "top_dir": dest_path,
            "extra_mapping_src_owner": [
                "linux-system-roles",
                "linux-system-roles",
                None,
            ],
            "extra_mapping_src_role": [
                otherroles[0],
                otherroles[1],
                otherroles[2],
            ],
            "extra_mapping_dest_prefix": [otherprefixdot, otherprefixdot, None],
            "extra_mapping_dest_role": [
                newotherroles[0],
                newotherroles[1],
                newotherroles[2],
            ],
        }
        copy_tree_with_replace(
            role_path,
            coll_path,
            rolename,
            rolename,
            MYTUPLE,
            transformer_args,
            isrole=True,
        )
        test_path = coll_path / "roles" / rolename / "tasks"
        print(coll_path)
        self.check_test_tree(
            test_path, test_yaml_str, post_params, ".yml", is_vertical=False
        )
        shutil.rmtree(coll_path)

    def test_import_replace(self):
        module_names = ["util0", "util1"]
        src_module_utils = []
        src_module_dir_core = Path(src_path) / rolename / "module_utils"
        src_module_dir_core.mkdir(parents=True, exist_ok=True)
        for mn in module_names:
            _src_module_util = src_module_dir_core / (mn + ".py")
            _src_module_util.touch(exist_ok=True)
            src_module_utils.append(_src_module_util)
        dest_base_dir = (
            Path(dest_path) / "ansible_collections" / namespace / collection_name
        )
        dest_module_dir_core = dest_base_dir / "plugins" / "module_utils"
        dest_module_dir_core.mkdir(parents=True, exist_ok=True)
        dest_module_utils = []
        for mn in module_names:
            _dest_module_util = dest_module_dir_core / mn
            _dest_module_util.touch(exist_ok=True)
            dest_module_utils.append(_dest_module_util)
        input = bytes(
            "import os\nimport ansible.module_utils.{0}  # noqa:E501\nimport ansible.module_utils.{1} as local1  # noqa:E501\nimport re\n".format(
                module_names[0], module_names[1]
            ),
            "utf-8",
        )
        expected = bytes(
            "import os\nimport ansible_collections.{0}.{1}.plugins.module_utils.{2}  # noqa:E501\nimport ansible_collections.{0}.{1}.plugins.module_utils.{3} as local1  # noqa:E501\nimport re\n".format(
                namespace, collection_name, module_names[0], module_names[1]
            ),
            "utf-8",
        )
        IMPORT_RE = re.compile(
            br"(\bimport) (ansible\.module_utils\.)(\S+)(.*)(\s+#.+|.*)$", flags=re.M
        )
        config["namespace"] = namespace
        config["collection"] = collection_name
        config["role"] = rolename
        config["src_path"] = Path(src_path) / rolename
        config["dest_path"] = dest_base_dir
        config["module_utils_dir"] = dest_module_dir_core
        config["module_utils"] = [
            [b"util0"],
            [b"util1"],
        ]
        config["additional_rewrites"] = []
        output = IMPORT_RE.sub(import_replace, input)
        self.assertEqual(output, expected)
        shutil.rmtree(src_module_dir_core)
        shutil.rmtree(dest_module_dir_core)

    def test_from_replace(self):
        module_name = "util0"
        src_module_dir = Path(src_path) / rolename / "module_utils" / module_name
        src_module_dir.mkdir(parents=True, exist_ok=True)
        dest_base_dir = (
            Path(dest_path) / "ansible_collections" / namespace / collection_name
        )
        dest_module_dir_core = dest_base_dir / "plugins" / "module_utils"
        dest_module_dir = dest_module_dir_core / module_name
        dest_module_dir.mkdir(parents=True, exist_ok=True)

        test_files = ["test0", "test1", "test2", "test3", "__init__"]
        for f in test_files:
            fpath = dest_module_dir / (f + ".py")
            fpath.touch(exist_ok=True)
        gather_module_utils_parts(dest_module_dir_core)

        input = bytes(
            textwrap.dedent(
                """\
                # pylint: disable=import-error, no-name-in-module
                {0}.basic import AnsibleModule
                {0}.{1} import {2}
                {0}.{1} import MyError
                {0}.{1}.{3} import (
                    ArgUtil,
                    ArgValidator_ListConnections,
                    ValidationError,
                )  # noqa:E501
                {0}.{1}.{4} import Util
                {0}.{1} import {5}
                """
            ).format(
                "from ansible.module_utils",
                module_name,
                test_files[0],
                test_files[1],
                test_files[2],
                test_files[3],
            ),
            "utf-8",
        )
        expected = bytes(
            textwrap.dedent(
                """\
                # pylint: disable=import-error, no-name-in-module
                from ansible.module_utils.basic import AnsibleModule
                {0}.{1}.{2}.plugins.module_utils.{3} import {4}
                {0}.{1}.{2}.plugins.module_utils.{3}.__init__ import MyError
                {0}.{1}.{2}.plugins.module_utils.{3}.{5} import (
                    ArgUtil,
                    ArgValidator_ListConnections,
                    ValidationError,
                )  # noqa:E501
                {0}.{1}.{2}.plugins.module_utils.{3}.{6} import Util
                {0}.{1}.{2}.plugins.module_utils.{3} import {7}
                """
            ).format(
                "from ansible_collections",
                namespace,
                collection_name,
                module_name,
                test_files[0],
                test_files[1],
                test_files[2],
                test_files[3],
            ),
            "utf-8",
        )
        FROM_RE = re.compile(
            br"(\bfrom) (ansible\.module_utils\.?)(\S+)? import (\(*(?:\n|\r\n)?)(.+)(\s+#.+|.*)$",
            flags=re.M,
        )
        config["namespace"] = namespace
        config["collection"] = collection_name
        config["role"] = rolename
        config["src_path"] = Path(src_path) / rolename
        config["dest_path"] = dest_base_dir
        config["module_utils_dir"] = dest_module_dir_core
        config["module_utils"] = [
            [b"util0"],
            [b"util0", b"test3"],
            [b"util0", b"test2"],
            [b"util0", b"test1"],
            [b"util0", b"test0"],
        ]
        config["additional_rewrites"] = []
        output = FROM_RE.sub(from_replace, input)
        self.assertEqual(output, expected)
        shutil.rmtree(src_module_dir)
        shutil.rmtree(dest_module_dir)

    def test_add_rolename(self):
        input = "README.md"
        expected = "README-" + rolename + ".md"
        output = add_rolename(input, rolename)
        self.assertEqual(output, expected)

        input = "README.test.md"
        expected = "README.test-" + rolename + ".md"
        output = add_rolename(input, rolename)
        self.assertEqual(output, expected)

        input = "LICENSE"
        expected = "LICENSE-" + rolename
        output = add_rolename(input, rolename)
        self.assertEqual(output, expected)
